"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint._typing import RuleMatches
from cfnlint.data.schemas.other import template as schema_template
from cfnlint.helpers import load_resource
from cfnlint.jsonschema import CfnTemplateValidator
from cfnlint.jsonschema._keywords_cfn import cfn_type
from cfnlint.rules.jsonschema.Base import BaseJsonSchema
from cfnlint.template import Template


class JsonSchema(BaseJsonSchema):
    """Check Base Template Settings"""

    id = "E1001"
    shortdesc = "Basic CloudFormation Template Configuration"
    description = (
        "Making sure the basic CloudFormation template components are properly"
        " configured"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["base"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.rule_set = {
            "cfnLint": "E1101",
            "condition": "E8007",
            "dynamicReference": "E1050",
            "fn_and": "E8004",
            "fn_base64": "E1021",
            "fn_cidr": "E1024",
            "fn_equals": "E8003",
            "fn_findinmap": "E1011",
            "fn_foreach": "E1032",
            "fn_getatt": "E1010",
            "fn_getazs": "E1015",
            "fn_if": "E1028",
            "fn_importvalue": "E1016",
            "fn_join": "E1022",
            "fn_length": "E1030",
            "fn_not": "E8005",
            "fn_or": "E8006",
            "fn_select": "E1017",
            "fn_split": "E1018",
            "fn_sub": "E1019",
            "fn_tojsonstring": "E1031",
            "format": "E1103",
            "ref": "E1020",
        }
        self.child_rules = dict.fromkeys(list(self.rule_set.values()))
        self.config_definition = {"sections": {"default": "", "type": "string"}}
        self.configure()
        self.validators = {
            "type": cfn_type,
        }

    @property
    def schema(self):
        schema = load_resource(schema_template, "configuration.json")
        if self.config["sections"]:
            schema["properties"][self.config["sections"]] = True

        return schema

    def match(self, cfn: Template) -> RuleMatches:
        """Check CloudFormation Parameters"""
        matches = []

        validator = CfnTemplateValidator({}).evolve(cfn=cfn)

        cfn_validator = self.extend_validator(
            validator=validator,
            schema=self.schema,
            context=cfn.context,
        )

        matches.extend(self.json_schema_validate(cfn_validator, cfn.template, []))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_lambda_eventsourcemapping
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class EventSourceMappingEventSourceArnStreamInclusive(CfnLintJsonSchema):
    id = "E3633"
    shortdesc = (
        "Validate Lambda event source mapping StartingPosition is used correctly"
    )
    description = (
        "When 'EventSourceArn' is associate to Kinesis, Kafka, "
        "or DynamoDB you must specify 'StartingPosition"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::Lambda::EventSourceMapping/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_lambda_eventsourcemapping,
                filename="eventsourcearn_stream_inclusive.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return "'StartingPosition' is a required property"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules import CloudFormationLintRule


class DynamicReferenceSsmPath(CloudFormationLintRule):
    id = "E1052"
    shortdesc = "Validate dynamic references to SSM are in a valid location"
    description = (
        "Dynamic references to SSM parameters are only supported "
        "in certain locations"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/dynamic-references.html#dynamic-references-ssm"
    tags = ["functions", "dynamic reference"]

    def validate(self, validator: Validator, s: Any, instance: Any, schema: Any):
        if len(validator.context.path.path) > 0:
            if validator.context.path.path[0] == "Parameters":
                if len(validator.context.path.path) >= 3:
                    if validator.context.path.path[2] in ["Default", "AllowedValues"]:
                        return
            elif validator.context.path.path[0] == "Resources":
                if len(validator.context.path.path) >= 3:
                    if validator.context.path.path[2] in ["Properties", "Metadata"]:
                        return
            elif validator.context.path.path[0] == "Outputs":
                if len(validator.context.path.path) >= 3:
                    if validator.context.path.path[2] in ["Value"]:
                        return

        yield ValidationError(
            (f"Dynamic reference {instance!r} to SSM parameters are not allowed here"),
            rule=self,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from unittest.mock import MagicMock, patch

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.properties.Properties import Properties
from cfnlint.schema._schema import Schema


@pytest.fixture(scope="module")
def rule():
    rule = Properties()
    yield rule


@pytest.mark.parametrize(
    "name,instance,patches,expected",
    [
        (
            "Valid results",
            {
                "Type": "MyType",
            },
            [(["us-east-1"], Schema({"typeName": "MyType", "properties": {}}))],
            [],
        ),
        (
            "Invalid type for Type",
            {
                "Type": {},
            },
            [],
            [],
        ),
        (
            "Valid type but no required fields",
            {
                "Type": "MyType",
            },
            [(["us-east-1"], Schema({"typeName": "MyType", "required": ["Name"]}))],
            [
                ValidationError(
                    "'Name' is a required property",
                    validator="required",
                    path=deque(["Properties"]),
                    schema_path=deque(["required"]),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, patches, expected, rule, validator):
    schema_manager = MagicMock()
    schema_manager.get_resource_schemas_by_regions.return_value = patches

    with patch(
        "cfnlint.rules.resources.properties.Properties.PROVIDER_SCHEMA_MANAGER",
        schema_manager,
    ):
        errs = list(rule.validate(validator, {}, instance, {}))

        assert errs == expected, f"Test {name!r} got {errs!r}"

    if patches:
        schema_manager.get_resource_schemas_by_regions.assert_called_once()
        schema_manager.get_resource_schemas_by_regions.assert_called_with(
            instance.get("Type"), ["us-east-1"]
        )
    else:
        schema_manager.get_resource_schemas_by_regions.assert_not_called()



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.templates.Description import Description


@pytest.fixture(scope="module")
def rule():
    rule = Description()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid description",
            "My Description",
            [],
        ),
        (
            "Invalid type",
            {},
            [
                ValidationError(
                    ("{} is not of type 'string'"),
                    rule=Description(),
                    schema_path=deque(["type"]),
                    validator="type",
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errors = list(rule.validate(validator, False, instance, {}))
    # we use error counts in this one as the instance types are
    # always changing so we aren't going to hold ourselves up by that
    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema
from __future__ import annotations

from collections import deque
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, ClassVar, Protocol, Type, runtime_checkable

from cfnlint.jsonschema._filter import FunctionFilter

if TYPE_CHECKING:
    from cfnlint.context import Context
    from cfnlint.jsonschema._format import FormatChecker
    from cfnlint.jsonschema._typing import ResolutionResult, V, ValidationResult
    from cfnlint.schema.resolver import RefResolver
    from cfnlint.template import Template


@runtime_checkable
class Validator(Protocol):
    """
    The protocol to which all validator classes adhere.

    Arguments:

        schema:

            The schema that the validator object will validate with.
            It is assumed to be valid, and providing
            an invalid schema can lead to undefined behavior.
    """

    #: A mapping of validation keywords (`str`\s) to functions that
    #: validate the keyword with that name. For more information see
    #: `creating-validators`.
    validators: ClassVar[Mapping]
    format_checker: ClassVar[FormatChecker]

    #: The schema that will be used to validate instances
    schema: Mapping | bool
    resolver: RefResolver

    cfn: Template
    context: Context
    function_filter: FunctionFilter
    cfn_path: deque[str]

    def __init__(
        self,
        schema: Mapping | bool,
    ) -> None: ...

    @classmethod
    def check_schema(cls, schema: Mapping | bool) -> None:
        """
        Validate the given schema against the validator's `META_SCHEMA`.

        Raises:

            `cfnlint.jsonschema.exceptions.SchemaError`:

                if the schema is invalid
        """

    def is_type(self, instance: Any, type: str) -> bool:
        """
        Check if the instance is of the given (JSON Schema) type.

        Arguments:

            instance:

                the value to check

            type:

                the name of a known (JSON Schema) type

        Returns:

            whether the instance is of the given type

        Raises:

            `cfnlint.jsonschema.exceptions.UnknownType`:

                if ``type`` is not a known type
        """

    def is_valid(self, instance: Any) -> bool:
        """
        Check if the instance is valid under the current `schema`.

        Returns:

            whether the instance is valid or not

        >>> schema = {"maxItems" : 2}
        >>> CfnTemplateValidator(schema).is_valid([2, 3, 4])
        False
        """

    def descend(
        self,
        instance: Any,
        schema: Any,
        path: str | int | None = None,
        schema_path: str | int | None = None,
        property_path: str | int | None = None,
    ) -> ValidationResult:
        """
        Descend into the schema validating the schema for True/False.
        It will validate the schema against the instance and append the
        path and schema_path as needed.

        >>> schema = {
        ...     "type" : "array",
        ...     "items" : {"enum" : [1, 2, 3]},
        ...     "maxItems" : 2,
        ... }
        >>> v = CfnTemplateValidator(schema)
        >>> for error in v.descend([2, 3, 4], schema, path=deque(["a"])):
        ...     print(error.message)
        """

    def iter_errors(self, instance: Any) -> ValidationResult:
        r"""
        Lazily yield each of the validation errors in the given instance.

        >>> schema = {
        ...     "type" : "array",
        ...     "items" : {"enum" : [1, 2, 3]},
        ...     "maxItems" : 2,
        ... }
        >>> v = CfnTemplateValidator(schema)
        >>> for error in sorted(v.iter_errors([2, 3, 4]), key=str):
        ...     print(error.message)
        4 is not one of [1, 2, 3]
        [2, 3, 4] is too long

        .. deprecated:: v4.0.0

            Calling this function with a second schema argument is deprecated.
            Use `Validator.evolve` instead.
        """

    def validate(self, instance: Any) -> None:
        """
        Check if the instance is valid under the current `schema`.

        Raises:

            `jsonschema.exceptions.ValidationError`:

                if the instance is invalid

        >>> schema = {"maxItems" : 2}
        >>> CfnTemplateValidator(schema).validate([2, 3, 4])
        Traceback (most recent call last):
            ...
        ValidationError: [2, 3, 4] is too long
        """

    def resolve_value(self, instance: Any) -> ResolutionResult:
        """
        Resolve the given instance, yielding each of its values.

        """

    def evolve(self, **kwargs) -> Validator:
        """
        Create a new validator like this one, but with given changes.

        Preserves all other attributes, so can be used to e.g. create a
        validator with a different schema but with the same :kw:`$ref`
        resolution behavior.

        >>> validator = CfnTemplateValidator({})
        >>> validator.evolve(schema={"type": "number"})
        CfnTemplateValidator(schema={'type': 'number'}, format_checker=None)

        The returned object satisfies the validator protocol, but may not
        be of the same concrete class! In particular this occurs
        when a :kw:`$ref` occurs to a schema with a different
        :kw:`$schema` than this one (i.e. for a different draft).

        >>> validator.evolve(
        ...     schema={"$schema": Draft7Validator.META_SCHEMA["$id"]}
        ... )
        Draft7Validator(schema=..., format_checker=None)
        """

    def extend(
        self,
        validators: dict[str, V] | None,
        function_filter: FunctionFilter | None = None,
        context: Context | None = None,
    ) -> Type[Validator]:
        """
        Extends the validator with a new set of validators to replace VALIDATORS.

        This function recall create with the a merged list of validators from the
        current instance merged with the parameters.

        >>> validator = CfnTemplateValidator({})
        >>> validator.extend(validators={"type": type})
        CfnTemplateValidator(schema={'type': 'number'}, format_checker=None)
        """



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_cloudwatch_alarm
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class AlarmAwsNamespacePeriod(CfnLintJsonSchema):
    id = "E3615"
    shortdesc = "Validate CloudWatch Alarm using AWS metrics has a correct period"
    description = (
        "Period < 60 not supported for namespaces with the following prefix: AWS/"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::CloudWatch::Alarm/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_cloudwatch_alarm,
                filename="aws_namespace_period.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return err.message



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.rds.DbClusterServerlessExclusive import (
    DbClusterServerlessExclusive,
)


@pytest.fixture(scope="module")
def rule():
    rule = DbClusterServerlessExclusive()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "EngineMode": "serverless",
            },
            [],
        ),
        (
            {
                "EngineMode": "Serverless",
            },
            [],
        ),
        (
            {"EngineMode": "serverless", "ServerlessV2ScalingConfiguration": "foo"},
            [
                ValidationError(
                    (
                        "EngineMode 'serverless'  doesn't allow additional "
                        "properties 'ServerlessV2ScalingConfiguration'"
                    ),
                    rule=DbClusterServerlessExclusive(),
                    path=deque(["ServerlessV2ScalingConfiguration"]),
                    validator=None,
                    schema_path=deque(
                        [
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "ServerlessV2ScalingConfiguration",
                        ]
                    ),
                )
            ],
        ),
        (
            {"EngineMode": "provisioned", "ScalingConfiguration": "foo"},
            [
                ValidationError(
                    (
                        "EngineMode 'provisioned'  doesn't allow "
                        "additional properties 'ScalingConfiguration'"
                    ),
                    rule=DbClusterServerlessExclusive(),
                    path=deque(["ScalingConfiguration"]),
                    validator=None,
                    schema_path=deque(
                        ["allOf", 0, "then", "properties", "ScalingConfiguration"]
                    ),
                )
            ],
        ),
        (
            {"ServerlessV2ScalingConfiguration": "foo"},
            [],
        ),
        (
            {"ScalingConfiguration": "foo"},
            [
                ValidationError(
                    (
                        "Additional properties are not allowed "
                        "('ScalingConfiguration')"
                    ),
                    rule=DbClusterServerlessExclusive(),
                    path=deque(["ScalingConfiguration"]),
                    validator=None,
                    schema_path=deque(
                        ["allOf", 2, "then", "properties", "ScalingConfiguration"]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.conditions.Condition import Condition


@pytest.fixture(scope="module")
def rule():
    rule = Condition()
    yield rule


@pytest.fixture
def template():
    return {
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]}
        },
    }


@pytest.fixture
def path():
    return Path(path=deque(["Resources", "MyRdsDbInstance", "Properties"]))


@pytest.mark.parametrize(
    "name,instance,errors",
    [
        ("Valid with string", {"Condition": "IsUsEast1"}, []),
        (
            "Invalid Type",
            {"Condition": []},
            [
                ValidationError(
                    "[] is not of type 'string'",
                    validator="condition",
                    schema_path=deque(["type"]),
                    path=deque(["Condition"]),
                ),
                ValidationError(
                    "[] is not one of ['IsUsEast1']",
                    validator="condition",
                    schema_path=deque(["enum"]),
                    path=deque(["Condition"]),
                ),
            ],
        ),
        (
            "Non existent condition",
            {"Condition": "IsUsWest2"},
            [
                ValidationError(
                    "'IsUsWest2' is not one of ['IsUsEast1']",
                    validator="condition",
                    schema_path=deque(["enum"]),
                    path=deque(["Condition"]),
                )
            ],
        ),
    ],
)
def test_condition(name, instance, errors, rule, validator):
    errs = list(rule.validate(validator, {}, instance, {}))

    assert errs == errors, name



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase


class TestPropertyOnlyOne(BaseRuleTestCase):
    """Test OnlyOne Property Configuration"""



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any, Sequence

from cfnlint.jsonschema import ValidationError
from cfnlint.rules import CloudFormationLintRule


class CfnLintKeyword(CloudFormationLintRule):
    def __init__(self, keywords: Sequence[str] | None = None) -> None:
        super().__init__()
        self.keywords = keywords or []
        self.parent_rules = ["E1101"]

    def message(self, instance: Any, err: ValidationError) -> str:
        return self.shortdesc



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.route53.RecordSet import RecordSet


@pytest.fixture(scope="module")
def rule():
    rule = RecordSet()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "Type": "A",
                "ResourceRecords": ["127.0.0.1"],
            },
            [],
        ),
        (
            {
                "Type": "A",
                "ResourceRecords": ["domain.local"],
            },
            [
                ValidationError(
                    "'domain.local' is not a 'ipv4'",
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 0]),
                    validator="format",
                    schema_path=deque(
                        [
                            "allOf",
                            0,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "format",
                        ]
                    ),
                )
            ],
        ),
        (
            {
                "Type": "TXT",
                "ResourceRecords": [
                    '"MS=ms123123"',
                    # 255 "a" characters + quotes
                    '"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"',
                    # 255 "a" characters and 255 "b" characters
                    '"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa""bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"',
                    # 255 "a" characters and 255 "b" characters and
                    # 255 "c" characters with a space between b and c
                    '"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa""bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb" "ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc"',  # noqa: E501
                ],
            },
            [],
        ),
        (
            {
                "Type": "TXT",
                "ResourceRecords": [
                    # 256 "a" characters - too long
                    '"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"',
                    # 256 "b" characters in a second record - too long
                    '"a""bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"',
                    '"MS=ms123123"',
                    # No quotation
                    "test2",
                ],
            },
            [
                ValidationError(
                    '\'"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"\' does not match \'^("[^"]{1,255}" *)*"[^"]{1,255}"$\'',  # noqa: E501
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 0]),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "allOf",
                            12,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "pattern",
                        ]
                    ),
                ),
                ValidationError(
                    '\'"a""bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"\' does not match \'^("[^"]{1,255}" *)*"[^"]{1,255}"$\'',  # noqa: E501
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 1]),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "allOf",
                            12,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "pattern",
                        ]
                    ),
                ),
                ValidationError(
                    '\'test2\' does not match \'^("[^"]{1,255}" *)*"[^"]{1,255}"$\'',
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 3]),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "allOf",
                            12,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "pattern",
                        ]
                    ),
                ),
            ],
        ),
        (
            {
                "Type": "AAAA",
                "ResourceRecords": [
                    "2001:0db8:85a3:0:0:8a2e:0370:7334",
                    "2001:0db8:3c4d:0015:0000:0000:1a2f:1a2b",
                    "0:0:0:0:0:0:A00:1",
                ],
            },
            [],
        ),
        (
            {
                "Type": "AAAA",
                "ResourceRecords": [
                    # IPv4 address
                    "127.0.0.1",
                    # Invalid data
                    "XXXX:XXXX:XXXX:XXXX:XXXX:XXXX:1.2.3.4",
                    # Spaces
                    "2001:0db8: 85a3:0:0:8a2e:0370:7334",
                    # Missing components
                    "1111:2222:3333:4444:5555:6666:7777",
                    # Too much components
                    "1111:2222:3333:4444:5555:6666:7777:8888::",
                ],
            },
            [
                ValidationError(
                    "'127.0.0.1' is not a 'ipv6'",
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 0]),
                    validator="format",
                    schema_path=deque(
                        [
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "format",
                        ]
                    ),
                ),
                ValidationError(
                    "'XXXX:XXXX:XXXX:XXXX:XXXX:XXXX:1.2.3.4' is not a 'ipv6'",
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 1]),
                    validator="format",
                    schema_path=deque(
                        [
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "format",
                        ]
                    ),
                ),
                ValidationError(
                    "'2001:0db8: 85a3:0:0:8a2e:0370:7334' is not a 'ipv6'",
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 2]),
                    validator="format",
                    schema_path=deque(
                        [
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "format",
                        ]
                    ),
                ),
                ValidationError(
                    "'1111:2222:3333:4444:5555:6666:7777' is not a 'ipv6'",
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 3]),
                    validator="format",
                    schema_path=deque(
                        [
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "format",
                        ]
                    ),
                ),
                ValidationError(
                    "'1111:2222:3333:4444:5555:6666:7777:8888::' is not a 'ipv6'",
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 4]),
                    validator="format",
                    schema_path=deque(
                        [
                            "allOf",
                            1,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "format",
                        ]
                    ),
                ),
            ],
        ),
        (
            {
                "Type": "CAA",
                "ResourceRecords": [
                    '0 issue "amazontrust.com;"',
                    '0 issue "awstrust.com;"',
                    '0 issue "amazonaws.com;"',
                ],
            },
            [],
        ),
        (
            {
                "Type": "CAA",
                "ResourceRecords": [
                    # No 3 items
                    "127.0.0.1",
                    # Missing quotes around the value
                    "0 issue amazon.com; henk",
                ],
            },
            [
                ValidationError(
                    (
                        "'127.0.0.1' does not match "
                        "'^(0|128)\\\\s([a-zA-Z0-9]+)\\\\s(\".+\")$'"
                    ),
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 0]),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "allOf",
                            2,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "pattern",
                        ]
                    ),
                ),
                ValidationError(
                    (
                        "'0 issue amazon.com; henk' does not match "
                        "'^(0|128)\\\\s([a-zA-Z0-9]+)\\\\s(\".+\")$'"
                    ),
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 1]),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "allOf",
                            2,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "pattern",
                        ]
                    ),
                ),
            ],
        ),
        (
            {
                "Type": "CNAME",
                "ResourceRecords": [
                    "hostname.example.com",
                ],
            },
            [],
        ),
        (
            {
                "Type": "CNAME",
                "ResourceRecords": [
                    "_x2.acm-validations.aws.",
                ],
            },
            [],
        ),
        (
            {
                "Type": "CNAME",
                "ResourceRecords": [
                    "cname1.example.com",
                    "foo√bar",
                ],
            },
            [
                ValidationError(
                    "'foo√bar' is not valid under any of the given schemas",
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 1]),
                    validator="anyOf",
                    context=deque(
                        [
                            ValidationError(
                                (
                                    "'foo√bar' does not match "
                                    "'^[a-zA-Z0-9\\\\!\"\\\\#\\\\$\\\\%\\\\&\\\\\\'\\\\(\\\\)\\\\*\\\\+\\\\,-\\\\/\\\\:\\\\;\\\\<\\\\=\\\\>\\\\?\\\\@\\\\[\\\\\\\\\\\\]\\\\^\\\\_\\\\`\\\\{\\\\|\\\\}\\\\~\\\\.]+$'"
                                ),  # noqa: E501
                                path=deque([]),
                                validator="pattern",
                                schema_path=deque([0, "pattern"]),
                                rule=RecordSet(),
                            ),
                            ValidationError(
                                (
                                    "'foo√bar' does not match "
                                    "'^.*\\\\.acm-validations\\\\.aws\\\\.?$'"
                                ),
                                path=deque([]),
                                validator="pattern",
                                schema_path=deque([1, "pattern"]),
                                rule=RecordSet(),
                            ),
                        ]
                    ),
                    schema_path=deque(
                        [
                            "allOf",
                            3,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "anyOf",
                        ]
                    ),
                ),
                ValidationError(
                    "['cname1.example.com', 'foo√bar'] is too long (1)",
                    rule=RecordSet(),
                    path=deque(["ResourceRecords"]),
                    validator="maxItems",
                    schema_path=deque(
                        [
                            "allOf",
                            3,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "maxItems",
                        ]
                    ),
                ),
            ],
        ),
        (
            {
                "Type": "CNAME",
                "ResourceRecords": [
                    # No valid domain name
                    "No valid domain name"
                ],
            },
            [
                ValidationError(
                    (
                        "'No valid domain name' is not valid "
                        "under any of the given schemas"
                    ),
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 0]),
                    validator="anyOf",
                    context=deque(
                        [
                            ValidationError(
                                (
                                    "'No valid domain name' does not match "
                                    "'^[a-zA-Z0-9\\\\!\"\\\\#\\\\$\\\\%\\\\&\\\\\\'\\\\(\\\\)\\\\*\\\\+\\\\,-\\\\/\\\\:\\\\;\\\\<\\\\=\\\\>\\\\?\\\\@\\\\[\\\\\\\\\\\\]\\\\^\\\\_\\\\`\\\\{\\\\|\\\\}\\\\~\\\\.]+$'"  # noqa: E501
                                ),
                                path=deque([]),
                                validator="pattern",
                                schema_path=deque([0, "pattern"]),
                                rule=RecordSet(),
                            ),
                            ValidationError(
                                (
                                    "'No valid domain name' does not match "
                                    "'^.*\\\\.acm-validations\\\\.aws\\\\.?$'"
                                ),
                                path=deque([]),
                                validator="pattern",
                                schema_path=deque([1, "pattern"]),
                                rule=RecordSet(),
                            ),
                        ]
                    ),
                    schema_path=deque(
                        [
                            "allOf",
                            3,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "anyOf",
                        ]
                    ),
                ),
            ],
        ),
        (
            {
                "Type": "MX",
                "ResourceRecords": [
                    "0 mx.example.com",
                    "65535 mx2.example.com",
                ],
            },
            [],
        ),
        (
            {
                "Type": "MX",
                "ResourceRecords": [
                    "127.0.0.1",
                    "10 my domain",
                ],
            },
            [
                ValidationError(
                    (
                        "'127.0.0.1' does not match "
                        "'^([0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{1-3}|65[0-4][0-9]{1-2}|655[0-2][0-9]|6553[0-5])\\\\s[a-zA-Z0-9\\\\!\"\\\\#\\\\$\\\\%\\\\&\\\\\\'\\\\(\\\\)\\\\*\\\\+\\\\,-\\\\/\\\\:\\\\;\\\\<\\\\=\\\\>\\\\?\\\\@\\\\[\\\\\\\\\\\\]\\\\^\\\\_\\\\`\\\\{\\\\|\\\\}\\\\~\\\\.]+$'"
                    ),
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 0]),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "allOf",
                            5,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "pattern",
                        ]
                    ),
                ),
                ValidationError(
                    (
                        "'10 my domain' does not match "
                        "'^([0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{1-3}|65[0-4][0-9]{1-2}|655[0-2][0-9]|6553[0-5])\\\\s[a-zA-Z0-9\\\\!\"\\\\#\\\\$\\\\%\\\\&\\\\\\'\\\\(\\\\)\\\\*\\\\+\\\\,-\\\\/\\\\:\\\\;\\\\<\\\\=\\\\>\\\\?\\\\@\\\\[\\\\\\\\\\\\]\\\\^\\\\_\\\\`\\\\{\\\\|\\\\}\\\\~\\\\.]+$'"
                    ),
                    rule=RecordSet(),
                    path=deque(["ResourceRecords", 1]),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "allOf",
                            5,
                            "then",
                            "properties",
                            "ResourceRecords",
                            "items",
                            "pattern",
                        ]
                    ),
                ),
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.DependsOnObsolete import DependsOnObsolete


@pytest.fixture(scope="module")
def rule():
    rule = DependsOnObsolete()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "Parent1Bucket": {
                "Type": "AWS::S3::Bucket",
            },
            "Parent2Bucket": {
                "Type": "AWS::S3::Bucket",
            },
            "ChildBucket": {
                "Type": "AWS::S3::Bucket",
                "Properties": {
                    "BucketName": {"Ref": "Parent1Bucket"},
                },
            },
        },
    }


@pytest.fixture
def path():
    return Path(
        path=deque(["Resources", "ChildBucket", "DependsOn"]),
    )


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid depends on",
            "Parent2Bucket",
            [],
        ),
        (
            "Valid but wrong type",
            {},
            [],
        ),
        (
            "Obsolete depends on",
            "Parent1Bucket",
            [
                ValidationError(
                    (
                        "'Parent1Bucket' dependency already "
                        "enforced by a 'Ref' at "
                        "'Resources/ChildBucket/Properties/BucketName'"
                    ),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, False, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Obsolete depends on",
            "Parent1Bucket",
            [],
        ),
    ],
)
def test_validate_no_cfn_graph(name, instance, expected, rule, validator):

    validator.cfn.graph = None
    errs = list(rule.validate(validator, False, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.extensions.aws_redshift_cluster
from cfnlint.rules.jsonschema.CfnLintJsonSchema import SchemaDetails
from cfnlint.rules.jsonschema.CfnLintJsonSchemaRegional import CfnLintJsonSchemaRegional


class ClusterNodeTypeEnum(CfnLintJsonSchemaRegional):
    id = "E3667"
    shortdesc = "Validate RedShift cluster node type"
    description = (
        "Validates the RedShift instance types based on region "
        "and data gathered from the pricing APIs"
    )
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::Redshift::Cluster/Properties/NodeType"],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.extensions.aws_redshift_cluster,
                "nodetype_enum.json",
            ),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

from cfnlint import ConfigMixIn
from cfnlint.rules import CloudFormationLintRule, Match
from cfnlint.runner import Runner


class ErrorRule(CloudFormationLintRule):
    id = "E0000"
    description = "Error"


class WarningRule(CloudFormationLintRule):
    id = "W0000"
    description = "Warning"


class InformationRule(CloudFormationLintRule):
    id = "I0000"
    description = "Informational"


error_match = Match("", ErrorRule(), "", 0, 0, 0, 0)
warning_match = Match("", WarningRule(), "", 0, 0, 0, 0)
informational_match = Match("", InformationRule(), "", 0, 0, 0, 0)


class TestExitCodes(BaseTestCase):
    """Test Exit Codes"""

    def test_exit_code_zero(self):
        params = [
            [[], "informational", 0],
            [[error_match], "informational", 2],
            [[error_match], "warning", 2],
            [[error_match], "error", 2],
            [[error_match], "none", 0],
            [[error_match, warning_match], "informational", 6],
            [[error_match, warning_match], "warning", 6],
            [[error_match, warning_match], "error", 2],
            [[error_match, warning_match], "none", 0],
            [[warning_match, informational_match], "informational", 12],
            [[warning_match, informational_match], "warning", 4],
            [[warning_match, informational_match], "error", 0],
            [[warning_match, informational_match], "none", 0],
            [[error_match, informational_match], "informational", 10],
            [[error_match, informational_match], "warning", 2],
            [[error_match, informational_match], "error", 2],
            [[error_match, informational_match], "none", 0],
            [[error_match, warning_match, informational_match], "informational", 14],
            [[error_match, warning_match, informational_match], "warning", 6],
            [[error_match, warning_match, informational_match], "error", 2],
            [[error_match, warning_match, informational_match], "none", 0],
        ]

        for matches, level, exit_code in params:
            with self.subTest():
                runner = Runner(ConfigMixIn(non_zero_exit_code=level))
                match_names = [x.rule.description for x in matches]
                with self.assertRaises(SystemExit) as e:
                    runner._exit(matches)
                self.assertEqual(
                    e.exception.code,
                    exit_code,
                    (
                        f"{match_names} matches with {level} "
                        f"level should yield {exit_code}"
                    ),
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class And(BaseFn):
    """Check And Condition Function Logic"""

    id = "E8004"
    shortdesc = "Check Fn::And structure for validity"
    description = "Check Fn::And is a list of two elements"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-and"
    tags = ["functions", "and"]

    def __init__(self) -> None:
        super().__init__("Fn::And", ("boolean",))
        self.fn_and = self.validate

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        return {
            "type": "array",
            "minItems": 2,
            "maxItems": 10,
            "fn_items": {
                "functions": [
                    "Condition",
                    "Fn::Equals",
                    "Fn::Not",
                    "Fn::And",
                    "Fn::Or",
                ],
                "schema": {
                    "type": ["boolean"],
                },
            },
        }



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.route53.HealthCheckHealthCheckConfigTypeInclusive import (
    HealthCheckHealthCheckConfigTypeInclusive,
)


@pytest.fixture(scope="module")
def rule():
    rule = HealthCheckHealthCheckConfigTypeInclusive()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"Type": "CLOUDWATCH_METRIC", "AlarmIdentifier": "Foo"},
            [],
        ),
        (
            {
                "Type": {"Ref": "AWS::Region"},
            },
            [],
        ),
        (
            [],
            [],
        ),
        (
            {
                "Type": "CLOUDWATCH_METRIC",
            },
            [
                ValidationError(
                    "'AlarmIdentifier' is a required property",
                    rule=HealthCheckHealthCheckConfigTypeInclusive(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "required"]),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.context._conditions import Conditions


def test_conditions():
    with pytest.raises(ValueError):
        Conditions.create_from_instance([])



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import pathlib

from cfnlint.helpers import load_plugins


def test_rule_descriptions():
    root_dir = pathlib.Path(__file__).parent.parent.parent.parent / "src/cfnlint/rules"
    rules = load_plugins(
        str(root_dir),
    )

    descriptions = set()
    shortdesc = set()

    for rule in rules:
        if not rule.id:
            continue
        assert (
            rule.description not in descriptions
        ), f"Duplicate description {rule.description!r}"
        assert (
            rule.shortdesc not in shortdesc
        ), f"Duplicate shortdesc {rule.shortdesc!r}"
        descriptions.add(rule.description)
        shortdesc.add(rule.shortdesc)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.jsonschema._keywords import maxLength
from cfnlint.rules import CloudFormationLintRule


class LimitDescription(CloudFormationLintRule):
    """Check Template Description Size"""

    id = "E1003"
    shortdesc = "Validate the max size of a description"
    description = (
        "Check if the size of the template description is less than the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["description", "limits"]

    def __init__(self) -> None:
        super().__init__()
        self.child_rules = {
            "I1003": None,
        }

    def maxLength(
        self, validator: Validator, mL: int, instance: Any, schema: Any
    ) -> ValidationResult:
        if not validator.is_type(instance, "string"):
            return

        errors = list(maxLength(validator, mL, instance, schema))

        if errors:
            yield from iter(errors)
            return

        for child_rule in self.child_rules.values():
            if child_rule is not None:
                if hasattr(child_rule, "maxLength"):
                    yield from child_rule.maxLength(validator, mL, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

import regex as re

from cfnlint.helpers import REGEX_SUB_PARAMETERS, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class Sub(BaseFn):
    """Check if Sub values are correct"""

    id = "E1019"
    shortdesc = "Sub validation of parameters"
    description = "Making sure the sub function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub"]

    def __init__(self) -> None:
        super().__init__("Fn::Sub", ("string",), resolved_rule="W1031")
        self.sub_parameter_types = ["string", "integer", "number", "boolean"]
        self.child_rules.update(
            {
                "W1019": None,
                "W1020": None,
            }
        )
        self._functions = [
            "Fn::Base64",
            "Fn::FindInMap",
            "Fn::GetAtt",
            "Fn::GetAZs",
            "Fn::If",
            "Fn::ImportValue",
            "Fn::Join",
            "Fn::Select",
            "Fn::Sub",
            "Fn::ToJsonString",
            "Fn::Transform",
            "Ref",
        ]

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        return {
            "type": ["array", "string"],
            "minItems": 2,
            "maxItems": 2,
            "fn_items": [
                {
                    "schema": {"type": "string"},
                },
                {
                    "functions": self._functions,
                    "schema": {
                        "type": ["object"],
                        "patternProperties": {
                            "[a-zA-Z0-9]+": {
                                "type": ["string"],
                            }
                        },
                        "additionalProperties": False,
                    },
                },
            ],
        }

    def _clean_error(
        self, err: ValidationError, instance: Any, param: Any
    ) -> ValidationError:
        err.message = err.message.replace(f"{instance!r}", f"{param!r}")
        err.instance = param
        err.path = deque([self.fn.name])
        err.schema_path = deque([])
        err.validator = self.fn.py
        return err

    def _validate_string(
        self, validator: Validator, key: str, instance: Any, sub_values: dict[str, Any]
    ) -> ValidationResult:
        params = re.findall(REGEX_SUB_PARAMETERS, instance)
        validator = validator.evolve(
            context=validator.context.evolve(
                functions=self._functions,
                path=validator.context.path.descend(
                    path=key,
                ),
                strict_types=True,
            ),
            function_filter=validator.function_filter.evolve(
                add_cfn_lint_keyword=False,
            ),
        )
        for param in params:
            param = param.strip()
            if "." in param:
                for err in validator.descend(
                    instance={"Fn::GetAtt": param},
                    schema={"type": ["string"]},
                ):
                    yield self._clean_error(err, {"Fn::GetAtt": param}, param)
            else:
                if param not in sub_values:
                    for err in validator.descend(
                        instance={"Ref": param},
                        schema={"type": ["string"]},
                    ):
                        yield self._clean_error(err, {"Ref": param}, param)

    def fn_sub(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        validator = validator.evolve(
            context=validator.context.evolve(strict_types=True),
        )
        errs = list(super().validate(validator, s, instance, schema))
        if errs:
            yield from iter(errs)
            return

        key, value = self.key_value(instance)
        sub_values = {}
        if validator.is_type(value, "array"):
            sub_values = value[1].copy()
            for sub_k, sub_v in value[1].items():
                fn_k, fn_v = is_function(sub_v)
                if fn_k == "Ref" and fn_v == sub_k:
                    del sub_values[sub_k]
            value = value[0]

        errs = list(self._validate_string(validator, key, value, sub_values))
        if errs:
            yield from iter(errs)
            return

        # we know the structure is valid at this point
        # so any child rule doesn't have to revalidate it
        for _, rule in self.child_rules.items():
            if rule and hasattr(rule, "validate"):
                for err in rule.validate(validator, s, instance.get("Fn::Sub"), schema):
                    err.path.append("Fn::Sub")
                    err.rule = rule
                    yield err



from __future__ import annotations

from collections import UserDict
from typing import Any

from cfnlint.schema import PROVIDER_SCHEMA_MANAGER, AttributeDict, ResourceNotFoundError


class _ResourceDict(UserDict):
    def __init__(self, __dict: None = None) -> None:
        self._has_modules: bool = False
        super().__init__(__dict)
        self.data: dict[str, AttributeDict] = {}

    def __setitem__(self, key: str, item: AttributeDict) -> None:
        if key.endswith(".*"):
            self._has_modules = True
        return super().__setitem__(key, item)

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def __getitem__(self, key: str) -> AttributeDict:
        attr = self.data.get(key)
        if attr is not None:
            return attr
        if not self._has_modules:
            raise KeyError(key)
        l_key = ""  # longest key
        for k in self.data.keys():
            if not k.endswith(".*"):
                continue
            if key.startswith(k[:-2]):
                if len(k) > len(l_key):
                    l_key = k
        if l_key:
            attr = self.data.get(l_key)
            if attr is not None:
                return attr
        raise KeyError(key)


class GetAtts:
    def __init__(self, regions: list[str]) -> None:
        self._regions = regions
        self._getatts: dict[str, _ResourceDict] = {}

        for region in self._regions:
            self._getatts[region] = _ResourceDict()

    def add(self, resource_name: str, resource_type: str) -> None:
        for region in self._regions:
            if resource_name not in self._getatts[region]:
                if resource_type.endswith("::MODULE"):
                    self._getatts[region][f"{resource_name}.*"] = (
                        PROVIDER_SCHEMA_MANAGER.get_type_getatts(
                            resource_type=resource_type, region=region
                        )
                    )

                try:
                    self._getatts[region][resource_name] = (
                        PROVIDER_SCHEMA_MANAGER.get_type_getatts(
                            resource_type=resource_type, region=region
                        )
                    )
                except ResourceNotFoundError:
                    self._getatts[region][resource_name] = AttributeDict()
                    continue

    def json_schema(self, region: str) -> dict:
        schema: dict[str, list[Any]] = {"oneOf": []}
        schema_strings: dict[str, Any] = {
            "type": "string",
            "enum": [],
        }
        schema_array: dict[str, str | list[Any]] = {
            "type": "array",
            "items": [{"type": "string", "enum": []}, {"type": ["string", "object"]}],
            "allOf": [],
        }

        for resource_name, attributes in self._getatts[region].items():
            attr_enum = []
            for attribute in attributes:
                attr_enum.append(attribute)
                schema_strings["enum"].append(
                    f"{resource_name}.{attribute}"
                )  # type: ignore

            schema_array["items"][0]["enum"].append(resource_name)  # type: ignore
            schema_array["allOf"].append(  # type: ignore
                {
                    "if": {
                        "items": [
                            {"type": "string", "const": resource_name},
                            {"type": ["string", "object"]},
                        ]
                    },
                    "then": {
                        "if": {
                            "items": [
                                {"type": "string", "const": resource_name},
                                {"type": "string"},
                            ]
                        },
                        "then": {
                            "items": [
                                {"type": "string", "const": resource_name},
                                {"type": "string", "enum": attr_enum},
                            ]
                        },
                        "else": {
                            "items": [
                                {"type": "string", "const": resource_name},
                                {
                                    "type": "object",
                                    "properties": {"Ref": {"type": "string"}},
                                    "required": ["Ref"],
                                    "additionalProperties": False,
                                },
                            ]
                        },
                    },
                    "else": {},
                }
            )

        schema["oneOf"].append(schema_array)
        schema["oneOf"].append(schema_strings)
        return schema

    def match(self, region: str, getatt: str | list[str]) -> str:
        if isinstance(getatt, str):
            getatt = getatt.split(".", 1)

        if isinstance(getatt, list):
            if len(getatt) != 2:
                raise TypeError("Invalid GetAtt size")

            try:
                result = (
                    self._getatts.get(region, _ResourceDict())
                    .get(getatt[0], AttributeDict())
                    .get(getatt[1])
                )
                if result is None:
                    raise ValueError("Attribute for resource doesn't exist")
                return result  # type: ignore
            except ValueError as e:
                raise e

        else:
            raise TypeError("Invalid GetAtt structure")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.schemas.other.resources
import cfnlint.helpers
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.jsonschema._keywords import patternProperties
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class Configuration(CfnLintJsonSchema):
    """Check Base Resource Configuration"""

    id = "E3001"
    shortdesc = "Basic CloudFormation Resource Check"
    description = (
        "Making sure the basic CloudFormation resources are properly configured"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["resources"]

    def __init__(self):
        super().__init__(
            keywords=["Resources"],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.resources, "configuration.json"
            ),
            all_matches=True,
        )
        self.validators = {
            "maxProperties": None,
            "propertyNames": None,
            "patternProperties": self._pattern_properties,
        }
        self.rule_set = {
            "maxProperties": "E3010",
            "propertyNames": "E3011",
        }
        self.child_rules = dict.fromkeys(list(self.rule_set.values()))

    def _pattern_properties(
        self, validator: Validator, aP: Any, instance: Any, schema: Any
    ):
        # We have to rework pattern properties
        # to re-add the keyword or we will have an
        # infinite loop
        validator = validator.evolve(
            function_filter=validator.function_filter.evolve(
                add_cfn_lint_keyword=True,
            )
        )

        yield from patternProperties(validator, aP, instance, schema)

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: Any
    ) -> ValidationResult:

        cfn_validator = self.extend_validator(
            validator=validator,
            schema=self._schema,
            context=validator.context.evolve(),
        )

        yield from self._iter_errors(cfn_validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json

from cfnlint.formatters.base import BaseFormatter
from cfnlint.rules import Match


class JsonFormatter(BaseFormatter):
    """Json Formatter"""

    class CustomEncoder(json.JSONEncoder):
        """Custom Encoding for the Match Object"""

        # pylint: disable=E0202

        def default(self, o):
            if isinstance(o, Match):
                return {
                    "Id": o.id,
                    "ParentId": o.parent_id,
                    "Rule": {
                        "Id": o.rule.id,
                        "Description": o.rule.description,
                        "ShortDescription": o.rule.shortdesc,
                        "Source": o.rule.source_url,
                    },
                    "Location": {
                        "Start": {
                            "ColumnNumber": o.columnnumber,
                            "LineNumber": o.linenumber,
                        },
                        "End": {
                            "ColumnNumber": o.columnnumberend,
                            "LineNumber": o.linenumberend,
                        },
                        "Path": getattr(o, "path", None),
                    },
                    "Level": o.rule.severity.capitalize(),
                    "Message": o.message,
                    "Filename": o.filename,
                }
            return {f"__{o.__class__.__name__}__": o.__dict__}

    def print_matches(self, matches, rules, config):
        # JSON formatter outputs a single JSON object
        # Unused argument http://pylint-messages.wikidot.com/messages:w0613
        del rules

        return json.dumps(
            list(matches),
            indent=4,
            cls=self.CustomEncoder,
            sort_keys=True,
            separators=(",", ": "),
        )



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
import os
import sys
from copy import deepcopy
from typing import Any, Iterator, Sequence

import cfnlint.formatters
import cfnlint.maintenance
from cfnlint.config import ConfigMixIn, configure_logging
from cfnlint.decode.decode import decode
from cfnlint.rules import Match, Rules
from cfnlint.rules.errors import ParseError, TransformError
from cfnlint.schema import PROVIDER_SCHEMA_MANAGER
from cfnlint.template.template import Template

LOGGER = logging.getLogger(__name__)


def get_formatter(config: ConfigMixIn) -> cfnlint.formatters.BaseFormatter:
    """
    Get a formatter instance based on the specified configuration.

    Args:
        config (ConfigMixIn): The configuration object containing the format setting.

    Returns:
        cfnlint.formatters.BaseFormatter: A formatter instance based
        on the specified format.

    """
    fmt = config.format
    if fmt:
        if fmt == "quiet":
            return cfnlint.formatters.QuietFormatter()
        if fmt == "parseable":
            # pylint: disable=bad-option-value
            return cfnlint.formatters.ParseableFormatter()
        if fmt == "json":
            return cfnlint.formatters.JsonFormatter()
        if fmt == "junit":
            return cfnlint.formatters.JUnitFormatter()
        if fmt == "pretty":
            return cfnlint.formatters.PrettyFormatter()
        if fmt == "sarif":
            return cfnlint.formatters.SARIFFormatter()

    return cfnlint.formatters.Formatter()


class TemplateRunner:
    """
    Runs a set of rules against a CloudFormation template.

    Attributes:
        config (ConfigMixIn): The configuration object containing
        settings for the template scan.
        cfn (Template): The CloudFormation template object.
        rules (Rules): The set of rules to be applied to the template.

    Methods:
        _dedup(matches: Iterator[Match]) -> Iterator[Match]:
            Deduplicate a sequence of matches.
        run() -> Iterator[Match]:
            Run the rules against the CloudFormation template and
            yield the resulting matches.
        check_metadata_directives(matches: Iterator[Match]) -> Iterator[Match]:
            Filter matches based on metadata directives in the template.

    """

    def __init__(
        self,
        filename: str | None,
        template: dict[str, Any],
        config: ConfigMixIn,
        rules: Rules,
    ) -> None:
        """
        Initialize a new TemplateRunner instance.

        Args:
            filename (str | None): The filename of the CloudFormation template.
            template (dict[str, Any]): The CloudFormation template as a dictionary.
            config (ConfigMixIn): The configuration object containing
            settings for the template scan.
            rules (Rules): The set of rules to be applied to the template.
        """
        self.config = deepcopy(config)
        self.config.set_template_args(template)
        self.cfn = Template(filename, template, self.config.regions)
        self.rules = rules

    def _dedup(self, matches: Iterator[Match]) -> Iterator[Match]:
        """
        Deduplicate a sequence of matches.

        Args:
            matches (Iterator[Match]): The sequence of matches to be deduplicated.

        Yields:
            Match: The unique matches from the input sequence.
        """
        seen: list[Match] = []
        for match in matches:
            if match not in seen:
                seen.append(match)
                yield match

    def run(self) -> Iterator[Match]:
        """
        Run the rules against the CloudFormation template and
        yield the resulting matches.

        Yields:
            Match: The matches found by running the rules against the template.
        """
        LOGGER.info("Run scan of template %s", self.cfn.filename)
        matches = self.cfn.transform()
        if matches:
            if self.rules.is_rule_enabled(TransformError(), self.config):
                yield from iter(matches)
            return

        if self.cfn.template is not None:
            if self.config.build_graph:
                self.cfn.build_graph()
            yield from self._dedup(
                self.check_metadata_directives(
                    self.rules.run(
                        filename=self.cfn.filename, cfn=self.cfn, config=self.config
                    )
                )
            )

    def check_metadata_directives(self, matches: Iterator[Match]) -> Iterator[Match]:
        """
        Filter matches based on metadata directives in the template.

        Args:
            matches (Iterator[Match]): The sequence of matches to be filtered.

        Yields:
            Match: The matches that are not suppressed by metadata directives.
        """
        directives = self.cfn.get_directives()

        for match in matches:
            if match.rule.id not in directives:
                yield match
            else:
                for mandatory_check in self.config.mandatory_checks:
                    if match.rule.id.startswith(mandatory_check):
                        yield match
                        break
                else:
                    path = getattr(match, "path", None)
                    if path:
                        if len(path) >= 2:
                            if path[0] != "Resources":
                                yield match
                                continue
                            if path[1] not in directives[match.rule.id]:
                                yield match
                        else:
                            yield match


class Runner:
    """
    Runs a set of rules against one or more CloudFormation templates.

    Attributes:
        config (ConfigMixIn): The configuration object containing
        settings for the template scan.
        formatter (BaseFormatter): The formatter used to output the
        results of the template scan.
        rules (Rules): The set of rules to be applied to the templates.

    Methods:
        _get_rules() -> None:
            Load the rules to be applied to the templates.
        _validate_filenames(filenames: Sequence[str | None]) -> Iterator[Match]:
            Validate the specified filenames and yield any matches found.
        validate_template(filename: str | None,
         template: dict[str, Any]) -> Iterator[Match]:
            Validate a single CloudFormation template and yield any matches found.
        _cli_output(matches: list[Match]) -> None:
            Output the results of the template scan to the console or a file.
        _exit(matches: list[Match]) -> int:
            Determine the appropriate exit code based on the severity of the matches.
        run() -> Iterator[Match]:
            Run the template validation process and yield any matches found.
        cli() -> None:
            Run the template validation process and output the results.
    """

    def __init__(self, config: ConfigMixIn) -> None:
        """
        Initialize a new Runner instance.

        Args:
            config (ConfigMixIn): The configuration object containing
                settings for the template scan.
        """
        self.config = config
        self.formatter = get_formatter(self.config)
        self.rules: Rules = Rules()
        self._get_rules()
        # load registry schemas before patching
        if self.config.registry_schemas:
            for path in self.config.registry_schemas:
                PROVIDER_SCHEMA_MANAGER.load_registry_schemas(path)
        # now we can patch after loading all registry schemas
        if self.config.override_spec:
            PROVIDER_SCHEMA_MANAGER.patch(
                self.config.override_spec, self.config.regions
            )

    def _get_rules(self) -> None:
        """
        Load the rules to be applied to the templates.
        """
        self.rules = Rules()
        try:
            for rules_path in self.config.append_rules:
                if rules_path and os.path.isdir(os.path.expanduser(rules_path)):
                    self.rules.update(Rules.create_from_directory(rules_path))
                else:
                    self.rules.update(Rules.create_from_module(rules_path))

            self.rules.update(
                Rules.create_from_custom_rules_file(self.config.custom_rules)
            )
        except (OSError, ImportError) as e:
            raise UnexpectedRuleException(
                f"Tried to append rules but got an error: {str(e)}", 1
            ) from e

    def _validate_filenames(self, filenames: Sequence[str | None]) -> Iterator[Match]:
        """
        Validate the specified filenames and yield any matches found.

        This function processes each filename in the provided sequence, decoding the
        template and validating it against the configured rules. Any matches found
        are yielded as an iterator.

        Args:
            filenames (Sequence[str | None]): The sequence of filenames to be validated.

        Yields:
            Match: The matches found during the validation process.

        Raises:
            None: This function does not raise any exceptions.
        """
        ignore_bad_template: bool = False
        if self.config.ignore_bad_template:
            ignore_bad_template = True
        else:
            # There is no collection at this point so we need to handle this
            # check directly
            if not ParseError().is_enabled(
                include_experimental=False,
                ignore_rules=self.config.ignore_checks,
                include_rules=self.config.include_checks,
                mandatory_rules=self.config.mandatory_checks,
            ):
                ignore_bad_template = True
        for filename in filenames:
            (template, matches) = decode(filename)
            if matches:
                if ignore_bad_template or any(
                    "E0000".startswith(x) for x in self.config.ignore_checks
                ):
                    matches = [match for match in matches if match.rule.id != "E0000"]
                    if matches:
                        yield from iter(matches)
                    return
                else:
                    yield from iter(matches)
                    return
            yield from self.validate_template(filename, template)  # type: ignore[arg-type] # noqa: E501

    def validate_template(
        self, filename: str | None, template: dict[str, Any]
    ) -> Iterator[Match]:
        """
        Validate a single CloudFormation template and yield any matches found.

        This function takes a CloudFormation template as a dictionary and runs the
        configured rules against it. Any matches found are yielded as an iterator.

        Args:
            filename (str | None): The filename of the CloudFormation template, or
                `None` if the template is not associated with a file.
            template (dict[str, Any]): The CloudFormation template as a dictionary.

        Yields:
            Match: The matches found during the validation process.

        Raises:
            None: This function does not raise any exceptions.
        """
        runner = TemplateRunner(filename, template, self.config, self.rules)
        yield from runner.run()

    def _cli_output(self, matches: list[Match]) -> None:
        formatter = get_formatter(self.config)
        matches.sort(key=lambda x: (x.filename, x.linenumber, x.rule.id))
        output = formatter.print_matches(matches, self.rules, config=self.config)
        if output:
            if self.config.output_file:
                with open(self.config.output_file, "w") as output_file:
                    output_file.write(output)
            else:
                print(output)

        self._exit(matches)

    def _exit(self, matches: list[Match]) -> int:
        """Determine exit code"""

        exit_level: str = self.config.non_zero_exit_code or "informational"

        exit_levels: dict[str, list[str]] = {
            "informational": ["informational", "warning", "error"],
            "warning": ["warning", "error"],
            "error": ["error"],
            "none": [],
        }

        exit_code = 0
        for match in matches:
            if (
                match.rule.severity == "informational"
                and match.rule.severity in exit_levels[exit_level]
            ):
                exit_code = exit_code | 8
            elif (
                match.rule.severity == "warning"
                and match.rule.severity in exit_levels[exit_level]
            ):
                exit_code = exit_code | 4
            elif (
                match.rule.severity == "error"
                and match.rule.severity in exit_levels[exit_level]
            ):
                exit_code = exit_code | 2

        sys.exit(exit_code)

    def run(self) -> Iterator[Match]:
        """
        Run the rules against the CloudFormation template and yield
        the resulting matches.

        This function is responsible for executing the validation
        process against the CloudFormation template. It performs
        the following steps:

        1. Transform the template, if necessary, and yield any matches found.
        2. Build the template graph, if configured.
        3. Run the configured rules against the template and
        yield the resulting matches.
        4. Filter the matches based on any metadata directives in the template.
        5. Deduplicate the matches before yielding them.

        Yields:
            Match: The matches found by running the rules against the template.

        Raises:
            None: This function does not raise any exceptions.
        """
        if not sys.stdin.isatty() and not self.config.templates:
            yield from self._validate_filenames([None])
            return

        yield from self._validate_filenames(self.config.templates)

    def cli(self) -> None:
        """
        Run the template validation process and output the results.

        This function is the entry point for the command-line interface (CLI) of the
        CloudFormation linter.

        If no templates are specified and the input is a terminal, the function will
        print the usage information and exit with a non-zero status code.

        Returns:
            None: This function does not return a value.

        Raises:
            None: This function does not raise any exceptions.
        """
        # Add our logging configuration when running CLI
        configure_logging(self.config.debug, self.config.info)

        if self.config.update_specs:
            cfnlint.maintenance.update_resource_specs(self.config.force)
            sys.exit(0)

        if self.config.update_iam_policies:
            cfnlint.maintenance.update_iam_policies()
            sys.exit(0)

        if self.config.update_documentation:
            # Get ALL rules (ignore the CLI settings))
            cfnlint.maintenance.update_documentation(self.rules)
            sys.exit(0)

        if self.config.listrules:
            print(self.rules)
            sys.exit(0)

        if not self.config.templates:
            if sys.stdin.isatty():
                self.config.parser.print_help()
                sys.exit(1)

        self._cli_output(list(self.run()))


def main() -> None:
    config = ConfigMixIn(sys.argv[1:])
    runner = Runner(config)
    runner.cli()


class CfnLintExitException(Exception):
    """
    An exception that is raised to indicate that the CloudFormation linter should exit.

    This exception is used to signal that the linter should exit
    with a specific exit code, typically indicating the severity
    of the issues found in the CloudFormation template.

    Attributes:
        exit_code (int): The exit code to be used when the linter exits.

    Methods:
        __init__(self, exit_code: int) -> None:
            Initialize a new CfnLintExitException instance with the specified exit code.
    """

    def __init__(self, msg=None, exit_code=1):
        """
        Initialize a new CfnLintExitException instance with the specified exit code.

        Args:
            exit_code (int): The exit code to be used when the linter exits.
        """
        if msg is None:
            msg = f"process failed with exit code {exit_code}"
        super().__init__(msg)
        self.exit_code = exit_code


class InvalidRegionException(CfnLintExitException):
    """
    An exception that is raised when an invalid AWS region is encountered.

    This exception is raised when the CloudFormation linter encounters a resource
    or parameter that references an AWS region that is not valid or supported.
    """


class UnexpectedRuleException(CfnLintExitException):
    """
    An exception that is raised when an unexpected error occurs while loading rules.

    This exception is raised when the CloudFormation linter encounters an error
    while attempting to load custom rules or rules from a specified directory or
    module. This could be due to a variety of reasons, such as a missing file,
    a syntax error in the rule code, or an issue with the rule implementation.
    """



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import sys


class color:
    error = "\033[31m"
    warning = "\033[33m"
    informational = "\033[34m"
    unknown = "\033[37m"
    green = "\033[32m"
    reset = "\033[0m"
    bold_reset = "\033[1:0m"
    underline_reset = "\033[4m"


def colored(s, c):
    """Takes in string s and outputs it with color"""
    if sys.stdout.isatty():
        return f"{c}{s}{color.reset}"

    return s



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class RequiredXor(CloudFormationLintRule):
    """Check Required Resource Configuration"""

    id = "E3014"
    shortdesc = "Validate only one of a set of required properties are specified"
    description = (
        "Make sure that Resources properties that are required exist. "
        "Along with other properties not being specified"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#requiredxor"
    tags = ["resources"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint._typing import RuleMatches
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class PreviousGenerationInstanceType(CloudFormationLintRule):
    id = "I3100"
    shortdesc = "Checks for legacy instance type generations"
    description = "New instance type generations increase performance and decrease cost"
    source_url = "https://aws.amazon.com/ec2/previous-generation/"
    tags = ["resources", "ec2", "rds", "elasticcache", "elasticsearch"]

    def match(self, cfn: Template) -> RuleMatches:
        matches = []
        for resource_type, property_type in [
            ("AWS::AutoScaling::LaunchConfiguration", "InstanceType"),
            ("AWS::EC2::CapacityReservation", "InstanceType"),
            ("AWS::EC2::Host", "InstanceType"),
            ("AWS::EC2::Instance", "InstanceType"),
            ("AWS::RDS::DBInstance", "DBInstanceClass"),
            ("AWS::ElastiCache::CacheCluster", "CacheNodeType"),
            ("AWS::ElastiCache::GlobalReplicationGroup", "CacheNodeType"),
            ("AWS::ElastiCache::ReplicationGroup", "CacheNodeType"),
        ]:
            for resource_name, resource in cfn.get_resources([resource_type]).items():
                if isinstance(
                    resource.get("Properties", {}).get(property_type, ""), str
                ):
                    if self.__is_previous_generation_instance_type(
                        resource.get("Properties", {}).get(property_type, ""),
                    ):
                        matches.append(
                            RuleMatch(
                                [
                                    "Resources",
                                    resource_name,
                                    "Properties",
                                    property_type,
                                ],
                                "Upgrade previous generation instance type: "
                                + resource.get("Properties", {}).get(property_type),
                            )
                        )

        for resource_type, top_level_property_type, property_type in [
            (
                "AWS::EC2::EC2Fleet",
                "FleetLaunchTemplateOverridesRequest",
                "InstanceType",
            ),
            ("AWS::EC2::LaunchTemplate", "LaunchTemplateData", "InstanceType"),
            ("AWS::EC2::SpotFleet", "SpotFleetLaunchSpecification", "InstanceType"),
            ("AWS::OpenSearchService::Domain", "ClusterConfig", "InstanceType"),
            (
                "AWS::Elasticsearch::Domain",
                "ElasticsearchClusterConfig",
                "InstanceType",
            ),
        ]:
            for resource_name, resource in cfn.get_resources([resource_type]).items():
                if isinstance(
                    resource.get("Properties", {})
                    .get(top_level_property_type, {})
                    .get(property_type, ""),
                    str,
                ):
                    if self.__is_previous_generation_instance_type(
                        resource.get("Properties", {})
                        .get(top_level_property_type, {})
                        .get(property_type, ""),
                    ):
                        matches.append(
                            RuleMatch(
                                [
                                    "Resources",
                                    resource_name,
                                    "Properties",
                                    top_level_property_type,
                                    property_type,
                                ],
                                "Upgrade previous generation instance type: "
                                + resource.get("Properties", {})
                                .get(top_level_property_type, {})
                                .get(property_type),
                            )
                        )
        return matches

    def __is_previous_generation_instance_type(self, instance_type):
        return (
            re.search(
                r"(^|\.)([cmr][1-3]|cc2|cg1|cr1|g2|hi1|hs1|i2|t1)($|\.)", instance_type
            )
            is not None
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_ec2_securitygroup
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class SecurityGroupAllToAndFromPorts(CfnLintJsonSchema):
    id = "E3688"
    shortdesc = "Validate that to and from ports are both -1"
    description = "When ToPort or FromPort are -1 the other one must also be -1"
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::EC2::SecurityGroup/Properties/SecurityGroupIngress/*",
                "Resources/AWS::EC2::SecurityGroup/Properties/SecurityGroupEgress/*",
                "Resources/AWS::EC2::SecurityGroupEgress/Properties",
                "Resources/AWS::EC2::SecurityGroupIngress/Properties",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_ec2_securitygroup,
                filename="all_to_and_from_ports.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return "Both ['FromPort', 'ToPort'] must be -1 when one is -1"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_rds_dbcluster
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class DbClusterSnapshotIdentifier(CfnLintJsonSchema):
    id = "W3688"
    shortdesc = "When restoring DBCluster certain properties are ignored"
    description = (
        "When creating a DBCluster extra properties are ignored and could "
        "result in drift"
    )
    tags = ["resources", "rds"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::RDS::DBCluster/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_rds_dbcluster,
                filename="snapshotidentifier.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return "Additional properties are ignored ('MasterUsername')"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from unittest import TestCase

from cfnlint.conditions._equals import Equal
from cfnlint.conditions._utils import get_hash


class TestEquals(TestCase):
    """Test Equals"""

    def setUp(self) -> None:
        self.equals = {"Ref": "AWS::Region"}
        self.equals_h = get_hash(self.equals)

        self.parameter_region = {"Ref": "Region"}
        self.parameter_region_h = get_hash(self.parameter_region)

        self.parameter_env = {"Ref": "Environment"}
        self.parameter_env_h = get_hash(self.parameter_env)

        return super().setUp()

    def test_equals_error_test(self):
        """Test equals scenarios condition"""
        equal = Equal([self.equals, "us-east-1"])

        with self.assertRaises(ValueError):
            equal.test({"foo": "bar"})

    def test_equals_left_test(self):
        """Test equals scenarios condition"""
        equal = Equal([self.equals, "us-east-1"])

        self.assertTrue(equal.test({self.equals_h: "us-east-1"}))
        self.assertFalse(equal.test({self.equals_h: "us-west-2"}))

        equal = Equal([self.equals, self.parameter_region])
        self.assertTrue(
            equal.test(
                {
                    self.equals_h: self.parameter_region_h,
                    self.parameter_region_h: self.equals_h,
                }
            )
        )
        self.assertFalse(equal.test({self.equals_h: self.parameter_env_h}))

    def test_equals_left_right(self):
        """Test equals scenarios condition"""
        equal = Equal(["us-east-1", self.equals])

        self.assertTrue(equal.test({self.equals_h: "us-east-1"}))
        self.assertFalse(equal.test({self.equals_h: "us-west-2"}))

        equal = Equal([self.parameter_region, self.equals])
        self.assertTrue(
            equal.test(
                {
                    self.equals_h: self.parameter_region_h,
                    self.parameter_region_h: self.equals_h,
                }
            )
        )
        self.assertFalse(equal.test({self.equals_h: self.parameter_env_h}))

    def test_equal_string_test(self):
        """Test equals scenarios condition"""
        equal = Equal(["us-west-2", "us-east-1"])

        self.assertFalse(equal.test({"foo": "bar"}))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint._typing import RuleMatches
from cfnlint.helpers import VALID_PARAMETER_TYPES_LIST
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class Default(CloudFormationLintRule):
    """Check if Parameters are configured correctly"""

    id = "E2015"
    shortdesc = "Default value is within parameter constraints"
    description = (
        "Making sure the parameters have a default value inside AllowedValues,"
        " MinValue, MaxValue, AllowedPattern"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html"
    tags = ["parameters"]

    def check_allowed_pattern(self, allowed_value, allowed_pattern, path):
        """
        Check allowed value against allowed pattern
        """
        message = "Default should be allowed by AllowedPattern"
        try:
            if not re.match(allowed_pattern, str(allowed_value)):
                return [RuleMatch(path, message)]
        except re.error as ex:
            self.logger.debug(
                'Regex pattern "%s" isn\'t supported by Python: %s', allowed_pattern, ex
            )

        return []

    def check_min_value(self, allowed_value, min_value, path):
        """
        Check allowed value against min value
        """
        message = "Default should be equal to or higher than MinValue"

        if isinstance(allowed_value, int) and isinstance(min_value, int):
            if allowed_value < min_value:
                return [RuleMatch(path, message)]

        return []

    def check_max_value(self, allowed_value, max_value, path):
        """
        Check allowed value against max value
        """
        message = "Default should be less than or equal to MaxValue"

        if isinstance(allowed_value, int) and isinstance(max_value, int):
            if allowed_value > max_value:
                return [RuleMatch(path, message)]

        return []

    def check_allowed_values(self, allowed_value, allowed_values, path):
        """
        Check allowed value against allowed values
        """
        message = "Default should be a value within AllowedValues"

        if allowed_value not in allowed_values:
            return [RuleMatch(path, message)]

        return []

    def check_min_length(self, allowed_value, min_length, path):
        """
        Check allowed value against MinLength
        """
        message = "Default should have a length above or equal to MinLength"

        value = allowed_value if isinstance(allowed_value, str) else str(allowed_value)
        if isinstance(min_length, int):
            if len(value) < min_length:
                return [RuleMatch(path, message)]

        return []

    def check_max_length(self, allowed_value, max_length, path):
        """
        Check allowed value against MaxLength
        """
        message = "Default should have a length below or equal to MaxLength"

        value = allowed_value if isinstance(allowed_value, str) else str(allowed_value)
        if isinstance(max_length, int):
            if len(value) > max_length:
                return [RuleMatch(path, message)]

        return []

    def match_default_value(self, paramname, paramvalue, default_value):
        matches = []

        if default_value is not None:
            path = ["Parameters", paramname, "Default"]
            allowed_pattern = paramvalue.get("AllowedPattern")
            if allowed_pattern:
                matches.extend(
                    self.check_allowed_pattern(default_value, allowed_pattern, path)
                )
            min_value = paramvalue.get("MinValue")
            if min_value:
                matches.extend(self.check_min_value(default_value, min_value, path))
            max_value = paramvalue.get("MaxValue")
            if max_value is not None:
                matches.extend(self.check_max_value(default_value, max_value, path))
            allowed_values = paramvalue.get("AllowedValues")
            if allowed_values:
                matches.extend(
                    self.check_allowed_values(default_value, allowed_values, path)
                )
            min_length = paramvalue.get("MinLength")
            if min_length is not None:
                matches.extend(self.check_min_length(default_value, min_length, path))
            max_length = paramvalue.get("MaxLength")
            if max_length is not None:
                matches.extend(self.check_max_length(default_value, max_length, path))
        return matches

    def is_list(self, paramvalue):
        return paramvalue.get("Type") in VALID_PARAMETER_TYPES_LIST

    def match(self, cfn: Template) -> RuleMatches:
        matches = []

        for paramname, paramvalue in cfn.get_parameters_valid().items():
            param_cdl_matches = []
            param_matches = []
            default_value = paramvalue.get("Default")
            if default_value is not None and self.is_list(paramvalue):
                comma_delimited_default_values = [
                    x.strip() for x in default_value.split(",")
                ]
                for value in comma_delimited_default_values:
                    param_cdl_matches.extend(
                        self.match_default_value(paramname, paramvalue, value)
                    )

            if param_cdl_matches or not self.is_list(paramvalue):
                param_matches.extend(
                    self.match_default_value(paramname, paramvalue, default_value)
                )

            if param_matches:
                matches.extend(param_matches)
                matches.extend(param_cdl_matches)

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
from typing import Any

from cfnlint.helpers import load_resource
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema
from cfnlint.schema.resolver import RefResolver


# pylint: disable=unused-argument
def _scalar_or_array(validator: Validator, subschema: Any, instance: Any, schema: Any):
    if validator.is_type(instance, "array"):
        for index, i in enumerate(instance):
            yield from validator.descend(i, subschema, path=index)
    else:
        yield from validator.descend(instance, subschema)


class Policy(CfnLintJsonSchema):
    """Check IAM policies"""

    def __init__(
        self,
        keywords: list[str] | None = None,
        schema_name: str | None = None,
        schema_file: str | None = None,
    ):
        super().__init__(keywords)

        if schema_name and schema_file:
            policy_schema = load_resource(
                "cfnlint.data.schemas.other.iam", "policy.json"
            )
            self.identity_schema = load_resource(
                "cfnlint.data.schemas.other.iam", schema_file
            )
            store = {
                "policy": policy_schema,
                schema_name: self.identity_schema,
            }
            # To reduce reduntant schemas use a schema resolver
            # this is deprecated in 4.18 of jsonschema
            self.resolver = RefResolver.from_schema(self.identity_schema, store=store)

    # pylint: disable=unused-argument
    def validate(
        self,
        validator: Validator,
        policy_type: Any,
        policy: Any,
        schema: dict[str, Any],
    ) -> ValidationResult:
        # First time child rules are configured against the rule
        # so we can run this now
        if validator.is_type(policy, "string"):
            try:
                iam_validator = validator.extend(
                    validators={
                        "scalarOrArray": _scalar_or_array,
                    },
                )(schema=self.identity_schema).evolve(
                    context=validator.context.evolve(
                        functions=[],
                    ),
                    resolver=self.resolver,
                )
                policy = json.loads(policy)
            except json.JSONDecodeError:
                return
        else:
            iam_validator = validator.extend(
                validators={
                    "scalarOrArray": _scalar_or_array,
                },
            )(schema=self.identity_schema).evolve(
                cfn=validator.cfn,
                context=validator.context,
                resolver=self.resolver,
            )

        for err in iam_validator.iter_errors(policy):
            if not err.validator.startswith("fn_") and err.validator not in ["cfnLint"]:
                err.rule = self
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.formatters.base import BaseFormatter


class JUnitFormatter(BaseFormatter):
    """JUnit-style Reports"""

    def _failure_format(self, match):
        """Format output of a failure"""
        formatstr = "{0} at {1}:{2}:{3}"
        return formatstr.format(
            match.message, match.filename, match.linenumber, match.columnnumber
        )

    def print_matches(self, matches, rules, config):
        """Output all the matches"""

        try:
            from junit_xml import TestCase, TestSuite, to_xml_report_string
        except ImportError as e:
            raise ImportError("Missing optional dependencies junit") from e

        if not rules:
            return None

        test_cases = []
        for rule in rules.values():
            if rule.id not in rules.used_rules:
                if not rule.id:
                    continue
                test_case = TestCase(name=f"{rule.id} {rule.shortdesc}")

                if rule.experimental:
                    test_case.add_skipped_info(
                        message="Experimental rule - not enabled"
                    )
                else:
                    test_case.add_skipped_info(message="Ignored rule")
                test_cases.append(test_case)
            else:
                test_case = TestCase(
                    name=f"{rule.id} {rule.shortdesc}",
                    allow_multiple_subelements=True,
                    url=rule.source_url,
                )
                for match in matches:
                    if match.rule.id == rule.id:
                        test_case.add_failure_info(
                            message=self._failure_format(match),
                            failure_type=match.message,
                        )
                test_cases.append(test_case)

        test_suite = TestSuite("CloudFormation Lint", test_cases)

        return to_xml_report_string([test_suite], prettyprint=True)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.iam.PolicyVersion import PolicyVersion


@pytest.fixture(scope="module")
def rule():
    rule = PolicyVersion()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        ("Valid version", "2012-10-17", []),
        (
            "Invalid version",
            "2008-10-17",
            [
                ValidationError(
                    "IAM Policy Version should be updated to '2012-10-17'",
                    rule=PolicyVersion(),
                ),
            ],
        ),
        ("No error on invalid structure", {}, []),
    ],
)
def test_policy_version(name, instance, expected, rule, validator):
    errors = list(rule.validate(validator, {}, instance, {}))
    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.data.schemas.other import resources
from cfnlint.helpers import load_resource
from cfnlint.jsonschema import Validator
from cfnlint.rules import CloudFormationLintRule


class Tagging(CloudFormationLintRule):
    id = "E3024"
    shortdesc = "Validate tag configuration"
    description = (
        "Validates tag values to make sure they have unique keys "
        "and they follow pattern requirements"
    )
    source_url = "https://docs.aws.amazon.com/tag-editor/latest/userguide/tagging.html"
    tags = ["parameters", "resources", "tags"]

    def __init__(self) -> None:
        super().__init__()
        self._schema = load_resource(resources, "tagging.json")

    def tagging(self, validator: Validator, t: Any, instance: Any, schema: Any):
        if not t.get("taggable"):
            return

        validator = validator.evolve(
            function_filter=validator.function_filter.evolve(
                add_cfn_lint_keyword=False,
            )
        )
        for err in validator.descend(
            instance=instance,
            schema=self._schema,
        ):
            err.validator = "tagging"
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from unittest import mock

import pytest

from cfnlint.context import Path
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.properties.ReadOnly import ReadOnly
from cfnlint.schema.resolver import RefResolutionError


@pytest.fixture(scope="module")
def rule():
    yield ReadOnly()


@pytest.fixture(scope="module")
def validator():
    yield CfnTemplateValidator(
        schema={
            "properties": {"Id": {"type": "string"}},
            "type": "object",
            "readOnlyProperties": ["/properties/Id"],
            "typeName": "Test",
        }
    )


@pytest.mark.parametrize(
    "name,path,side_effect,expected",
    [
        (
            "Not using read only property",
            deque(["Resources", "Test", "Properties", "Name"]),
            None,
            [],
        ),
        (
            "Not in outputs",
            deque(["Outputs"]),
            None,
            [],
        ),
        (
            "Not in resource properties",
            deque(["Resources"]),
            None,
            [],
        ),
        (
            "Not in resource properties",
            deque(["Resources", "*", "Metadata"]),
            None,
            [],
        ),
        (
            "Setting a read only property",
            deque(["Resources", "Test", "Properties", "Id"]),
            None,
            [ValidationError("Read only properties are not allowed ('Id')")],
        ),
        (
            "No readonlyProperties in schema",
            deque(["Resources", "Test", "Properties", "Id"]),
            RefResolutionError("not found"),
            [],
        ),
    ],
)
def test_validate(name, path, side_effect, expected, rule, validator):
    validator = validator.evolve(
        context=validator.context.evolve(
            path=Path(
                cfn_path=path,
            )
        ),
    )

    if side_effect:
        validator.resolver = mock.MagicMock()
        validator.resolver.resolve_from_url.side_effect = side_effect

    errs = list(rule.validate(validator, "", "", {}))

    assert errs == expected, f"{name} got errors {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.CfnInit import CfnInit


@pytest.fixture(scope="module")
def rule():
    rule = CfnInit()
    yield rule


@pytest.fixture
def template():
    return {
        "Parameters": {
            "MyString": {
                "Type": "String",
            }
        }
    }


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid Metadata",
            {
                "config": {
                    "packages": {"yum": {"httpd": []}},
                    "files": {
                        "/var/www/html/index.html": {
                            "content": (
                                "<body>\n  <h1>Congratulations, you have "
                                "successfully launched the AWS "
                                "CloudFormation sample.</h1>\n</body>\n"
                            ),
                            "mode": "000644",
                            "owner": "root",
                            "group": "root",
                        },
                        "/etc/cfn/cfn-hup.conf": {
                            "content": {
                                "Fn::Sub": (
                                    "[main]\nstack="
                                    "${AWS::StackId}\n"
                                    "region=${AWS::Region}\n"
                                )
                            },
                            "mode": "000400",
                            "owner": "root",
                            "group": "root",
                        },
                        "/etc/cfn/hooks.d/cfn-auto-reloader.conf": {
                            "content": {
                                "Fn::Sub": (
                                    "[cfn-auto-reloader-hook]\ntriggers="
                                    "post.update\npath=Resources."
                                    "LaunchConfig.Metadata.AWS::CloudFormation::Init"
                                    "\naction=/opt/aws/bin/cfn-init "
                                    "-v --stack ${AWS::StackName}"
                                    " --resource Instance --region "
                                    "${AWS::Region}\nrunas=root"
                                )
                            }
                        },
                    },
                    "services": {
                        "sysvinit": {
                            "httpd": {"enabled": True, "ensureRunning": True},
                            "cfn-hup": {
                                "enabled": True,
                                "ensureRunning": True,
                                "files": [
                                    "/etc/cfn/cfn-hup.conf",
                                    "/etc/cfn/hooks.d/cfn-auto-reloader.conf",
                                ],
                            },
                        }
                    },
                },
            },
            [],
        ),
        (
            "Bad type",
            [],
            [
                ValidationError(
                    ("[] is not of type 'object'"),
                    path=deque([]),
                    schema_path=deque(["type"]),
                    validator="type",
                    rule=CfnInit(),
                    path_override=deque([]),
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint._typing import RuleMatches
from cfnlint.helpers import is_function
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class Used(CloudFormationLintRule):
    """Check if Mappings are used anywhere in the template"""

    id = "W7001"
    shortdesc = "Check if Mappings are Used"
    description = "Making sure the mappings defined are used"
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["mappings"]

    def match(self, cfn: Template) -> RuleMatches:
        matches: RuleMatches = []
        findinmap_mappings = []

        mappings = cfn.template.get("Mappings", {})
        k, _ = is_function(mappings)
        if k == "Fn::Transform":
            self.logger.debug(
                (f"Mapping Name has a transform. Disabling check {self.id!r}"),
            )
            return matches

        if mappings:
            # Get all "FindInMaps" that reference a Mapping
            maptrees = cfn.transform_pre["Fn::FindInMap"]
            for maptree in maptrees:
                if isinstance(maptree[-1], list):
                    map_name = maptree[-1][0]
                    if isinstance(map_name, dict):
                        self.logger.debug(
                            (
                                "Mapping Name has a function that can have too many"
                                " variations. Disabling check %s"
                            ),
                            self.id,
                        )
                        return matches

                    findinmap_mappings.append(maptree[-1][0])
                else:
                    findinmap_mappings.append(maptree[-1])

            # Check if the mappings are used
            for mapname, _ in mappings.items():
                if mapname not in findinmap_mappings:
                    message = "Mapping '{0}' is defined but not used"
                    matches.append(
                        RuleMatch(["Mappings", mapname], message.format(mapname))
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.schema.resolver._exceptions import RefResolutionError
from cfnlint.schema.resolver._resolver import RefResolver
from cfnlint.schema.resolver._utils import id_of

__all__ = [
    "RefResolver",
    "RefResolutionError",
    "id_of",
]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from unittest import TestCase

from cfnlint.context import Context
from cfnlint.jsonschema import StandardValidator, ValidationError
from cfnlint.rules.jsonschema.Base import BaseJsonSchema


class Rule(BaseJsonSchema):
    def __init__(self):
        """Init"""
        super().__init__()
        self.rule_set = {"enum": "E3003", "pattern": "E3004"}
        self.child_rules = dict.fromkeys(list(self.rule_set.values()))


class ReturnRule(BaseJsonSchema):
    id = "E3005"


class EnumRule(BaseJsonSchema):
    id = "E3003"

    def __init__(self) -> None:
        super().__init__()
        self.rule = ReturnRule()

    def enum(self, validator, enums, instance, schema):
        yield ValidationError("test", rule=self.rule)


class PatternRule(BaseJsonSchema):
    id = "E3004"

    def pattern(self, validator, ptrn, instance, schema):
        yield ValidationError("test")


class TestBaseJsonSchema(TestCase):
    """Test Base JSON Schema validation"""

    def setUp(self) -> None:
        super().setUp()
        self.rule = Rule()

    def test_rule_disabled(self):
        schema = {
            "type": "string",
            "enum": ["foo"],
        }
        validator = self.rule.extend_validator(
            validator=StandardValidator({}),
            schema=schema,
            context=Context(),
        )

        self.assertEqual(self.rule.json_schema_validate(validator, "bar", []), [])

    def test_rule_enabled(self):
        schema = {
            "type": "string",
            "pattern": "foo",
        }
        self.rule.child_rules["E3004"] = PatternRule()
        validator = self.rule.extend_validator(
            validator=StandardValidator({}),
            schema=schema,
            context=Context(),
        )

        self.assertEqual(
            len(list(self.rule.json_schema_validate(validator, "bar", []))), 1
        )
        self.assertEqual(
            list(self.rule.json_schema_validate(validator, "bar", []))[0].rule.id,
            "E3004",
        )

    def test_rule_passed_in_validation_error(self):
        schema = {
            "type": "string",
            "enum": ["foo"],
        }
        self.rule.child_rules["E3003"] = EnumRule()
        validator = self.rule.extend_validator(
            validator=StandardValidator({}),
            schema=schema,
            context=Context(),
        )

        self.assertEqual(
            list(self.rule.json_schema_validate(validator, "bar", []))[0].rule.id,
            "E3005",
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules import CloudFormationLintRule


class ApproachingLimitDescription(CloudFormationLintRule):
    """Check Template Description Size"""

    id = "I1003"
    shortdesc = "Validate if we are approaching the max size of a description"
    description = (
        "Check if the size of the template description is approaching the upper limit"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["description", "limits"]

    def __init__(self) -> None:
        super().__init__()
        self.config["threshold"] = 0.9

    def maxLength(
        self, validator: Validator, mL: int, instance: Any, schema: Any
    ) -> ValidationResult:
        if len(instance) > mL * self.config.get("threshold", 1):
            yield ValidationError(
                f"{instance!r} is approaching the max length of {mL}",
                rule=self,
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.Cidr import Cidr


@pytest.fixture(scope="module")
def rule():
    rule = Cidr()
    yield rule


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::Cidr with 2 element array",
            {"Fn::Cidr": ["192.168.0.0/24", 6]},
            {"type": "array"},
            [],
        ),
        (
            "Valid Fn::Cidr with 3 element array",
            {"Fn::Cidr": ["192.168.0.0/24", 6, 5]},
            {"type": "array"},
            [],
        ),
        (
            "Invalid Fn::Cidr with wrong output type",
            {"Fn::Cidr": ["192.168.0.0/24", 2]},
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::Cidr': ['192.168.0.0/24', 2]} is not of type 'string'",
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_cidr",
                ),
            ],
        ),
        (
            "Invalid Fn::Cidr is NOT a array",
            {"Fn::Cidr": "foo"},
            {"type": "array"},
            [
                ValidationError(
                    "'foo' is not of type 'array'",
                    path=deque(["Fn::Cidr"]),
                    schema_path=deque(["type"]),
                    validator="fn_cidr",
                ),
            ],
        ),
        (
            "Valid Fn::Cidr with a valid function",
            {"Fn::Cidr": ["192.168.0.0/24", {"Fn::FindInMap": ["A", "B", "D"]}]},
            {"type": "array"},
            [],
        ),
        (
            "Invalid Fn::Cidr with an invalid function",
            {"Fn::Cidr": ["192.168.0.0/24", {"Fn::Join": ["-", "bar"]}]},
            {"type": "array"},
            [
                ValidationError(
                    "{'Fn::Join': ['-', 'bar']} is not of type 'integer'",
                    path=deque(["Fn::Cidr", 1]),
                    schema_path=deque(["fn_items", "type"]),
                    validator="fn_cidr",
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.fn_cidr(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any, Dict

import jsonpatch

from cfnlint.schema._getatts import AttributeDict, GetAtts
from cfnlint.schema.resolver import RefResolver

# Can't use a dataclass because its hard to parse in json
# with optional fields without addtional help


class Schema:
    def __init__(self, schema: dict[str, Any], is_cached: bool = False) -> None:
        self.is_cached: bool = is_cached
        self._schema: dict[str, Any] = schema
        self._type_name: str = schema["typeName"]
        self.resolver: RefResolver = RefResolver.from_schema(schema)
        self._getatts: GetAtts = GetAtts(self)

    @property
    def type_name(self) -> str:
        """Return the type name of the schema"""
        return self._type_name

    @property
    def schema(self) -> Dict:
        """Return the JSON Schema version of the CFN Schema

        Args:
        Returns:
            Dict: the JSON schema representation of the CFN schema
        """
        return self._schema

    def patch(self, patches: list[Dict]) -> None:
        """Patches the schema file

        Args:
            patches: A list of JSON Patches
        Returns:
            None: Returns when the patches have been applied
        """
        jsonpatch.JsonPatch(patches).apply(self.schema, in_place=True)

    @property
    def get_atts(self) -> AttributeDict:
        """Get the valid GetAtts for this schema. Schemas are defined in property
            readOnlyProperties and we need to build definitions of those properties

        Args:
        Returns:
            dict[str, GetAtt]: Dict of keys with valid strings and the json schema
            object for the property
        """
        return self._getatts.attrs



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.certificatemanager.DomainValidationOptions import (
    DomainValidationOptions,
)


@pytest.fixture(scope="module")
def rule():
    rule = DomainValidationOptions()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "DomainName": "foo.bar",
                "ValidationDomain": "bar",
            },
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "DomainName": "foo.bar",
                "ValidationDomain": "foo.bar",
            },
            [],
        ),
        (
            {
                "DomainName": "foo.bar",
                "ValidationDomain": {"Ref": "pValidationDomain"},
            },
            [],
        ),
        (
            {
                "DomainName": {"Ref": "pDomainName"},
                "ValidationDomain": "bar",
            },
            [],
        ),
        (
            {
                "DomainName": "foobar",
                "ValidationDomain": "bar",
            },
            [
                ValidationError(
                    "'bar' must be a superdomain of 'foobar'",
                    path=deque(["DomainName"]),
                )
            ],
        ),
        (
            {
                "DomainName": "foo.foo",
                "ValidationDomain": "bar",
            },
            [
                ValidationError(
                    "'bar' must be a superdomain of 'foo.foo'",
                    path=deque(["DomainName"]),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.extensions.aws_elasticloadbalancingv2_listener
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class ListenerCertificate(CfnLintJsonSchema):
    id = "E3676"
    shortdesc = (
        "Validate ELBv2 protocols that require certificates "
        "have a certificate specified"
    )
    description = "When using HTTP or TLS you must specify a certificate"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-elasticloadbalancingv2-listener.html#cfn-elasticloadbalancingv2-listener-certificates"
    tags = ["resources"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::ElasticLoadBalancingV2::Listener/Properties",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_elasticloadbalancingv2_listener,
                filename="certificate.json",
            ),
            all_matches=True,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

import cfnlint.decode.cfn_json  # pylint: disable=E0401
import cfnlint.decode.cfn_yaml  # pylint: disable=E0401
from cfnlint import ConfigMixIn
from cfnlint.config import _DEFAULT_RULESDIR
from cfnlint.rules import Rules
from cfnlint.runner import TemplateRunner


class TestRunner(BaseTestCase):
    """Test Duplicates Parsing"""

    def setUp(self):
        """SetUp template object"""
        self.rules = Rules.create_from_directory(_DEFAULT_RULESDIR)
        (self.template, _) = cfnlint.decode.decode_str(
            """
            AWSTemplateFormatVersion: "2010-09-09"
            Description: >
                Template with all error levels: Warning, Error and Informational
            # Adding in an issue outside of the Resources for validating
            Conditions:
                IsUsEast1: !Equals ["a", "b", "c"]
            Resources:
                myTable:
                    Metadata:
                        cfn-lint:
                            config:
                                ignore_checks:
                                - I3011
                                - W1020
                                - E8003 # condition error #
                    Type: "AWS::DynamoDB::Table"
                    Properties:
                        TableName: !Sub "TableName"
                        AttributeDefinitions:
                            - AttributeName: "Id"
                              AttributeType: "S" # Valid AllowedValue
                        KeySchema:
                            - AttributeName: "Id"
                              KeyType: "HASH"
                        ProvisionedThroughput:
                            ReadCapacityUnits: 5
                            WriteCapacityUnits: "5"
                myTable2:
                    # With no ignore_checks so we
                    # should still get issues from this
                    Type: "AWS::DynamoDB::Table"
                    Properties:
                        TableName: !Sub "TableName1"
                        AttributeDefinitions:
                            - AttributeName: "Id"
                              AttributeType: "S" # Valid AllowedValue
                        KeySchema:
                            - AttributeName: "Id"
                              KeyType: "HASH"
                        ProvisionedThroughput:
                            ReadCapacityUnits: !If [IsUsEast1, 5, 5]
                            WriteCapacityUnits: "5"
        """
        )

    def test_runner(self):
        """Success test"""
        runner = TemplateRunner(
            filename=None,
            template=self.template,
            config=ConfigMixIn(
                regions=["us-east-1"],
                include_checks=["I"],
                include_experimental=True,
            ),
            rules=self.rules,
        )
        failures = list(runner.run())
        self.assertEqual(len(failures), 4, "Got failures {}".format(failures))

    def test_runner_mandatory_rules(self):
        """Success test"""
        runner = TemplateRunner(
            filename=None,
            template=self.template,
            config=ConfigMixIn(
                mandatory_checks=["W1020"],
                regions=["us-east-1"],
                include_checks=["I"],
                include_experimental=True,
            ),
            rules=self.rules,
        )
        failures = list(runner.run())
        self.assertEqual(len(failures), 5, "Got failures {}".format(failures))

        runner = TemplateRunner(
            filename=None,
            template=self.template,
            config=ConfigMixIn(
                mandatory_checks=["W9000"],
                regions=["us-east-1"],
                include_checks=["I"],
                include_experimental=True,
            ),
            rules=self.rules,
        )
        failures = list(runner.run())
        self.assertEqual(len(failures), 4, "Got failures {}".format(failures))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.parameters.Configuration import Configuration
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = Configuration()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {},
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Empty is okay",
            {},
            [],
        ),
        (
            "wrong type",
            [],
            [
                ValidationError(
                    "[] is not of type 'object'",
                    validator="type",
                    schema_path=deque(["type"]),
                    rule=Configuration(),
                )
            ],
        ),
        (
            "String type with Parameter",
            {
                "MyString": {
                    "Type": "String",
                    "AllowedPattern": "1",
                }
            },
            [],
        ),
        (
            "Number type with MinValue",
            {
                "MyNumber": {
                    "Type": "Number",
                    "MinValue": "1",
                }
            },
            [],
        ),
        (
            "AWS type allowed allowed pattern",
            {
                "AWS": {
                    "Type": "AWS::EC2::Image::Id",
                    "AllowedPattern": "^ami-[0-9a-f]+$",
                }
            },
            [],
        ),
        (
            "Number type with AllowedPattern",
            {
                "MyNumber": {
                    "Type": "Number",
                    "AllowedPattern": "1",
                }
            },
            [
                ValidationError(
                    (
                        "'AllowedPattern' is not one of ['AllowedValues', "
                        "'ConstraintDescription', 'Default', 'Description', "
                        "'MaxValue', 'MinValue', 'NoEcho', 'Type']"
                    ),
                    validator="enum",
                    schema_path=deque(
                        [
                            "patternProperties",
                            "^[a-zA-Z0-9]+$",
                            "then",
                            "propertyNames",
                            "enum",
                        ]
                    ),
                    rule=Configuration(),
                    path=deque(["MyNumber", "AllowedPattern"]),
                )
            ],
        ),
        (
            "String type with MinValue",
            {
                "MyString": {
                    "Type": "String",
                    "MinValue": "1",
                }
            },
            [
                ValidationError(
                    (
                        "'MinValue' is not one of ['AllowedPattern', "
                        "'AllowedValues', 'ConstraintDescription', "
                        "'Default', 'Description', 'MaxLength', "
                        "'MinLength', 'NoEcho', 'Type']"
                    ),
                    validator="enum",
                    schema_path=deque(
                        [
                            "patternProperties",
                            "^[a-zA-Z0-9]+$",
                            "else",
                            "then",
                            "propertyNames",
                            "enum",
                        ]
                    ),
                    rule=Configuration(),
                    path=deque(["MyString", "MinValue"]),
                )
            ],
        ),
        (
            "Bad Name",
            {
                "Ref&3": {
                    "Type": "String",
                }
            },
            [
                ValidationError(
                    "'Ref&3' does not match any of the regexes: '^[a-zA-Z0-9]+$'",
                    validator="additionalProperties",
                    schema_path=deque(["additionalProperties"]),
                    rule=Configuration(),
                    path=deque(["Ref&3"]),
                )
            ],
        ),
        (
            "Long key name",
            {
                "a"
                * 256: {
                    "Type": "String",
                }
            },
            [
                ValidationError(
                    f"{'a'*256!r} is longer than 255",
                    validator="maxLength",
                    schema_path=deque(["propertyNames", "maxLength"]),
                    rule=None,  # none becuase we don't load child rule
                    path=deque(["a" * 256]),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, context, cfn):
    validator = CfnTemplateValidator(context=context, cfn=cfn)
    errs = list(rule.validate(validator, {}, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
import os
from pathlib import Path
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

import cfnlint.config  # pylint: disable=E0401
from cfnlint.helpers import REGIONS

LOGGER = logging.getLogger("cfnlint")


class TestConfigMixIn(BaseTestCase):
    """Test ConfigParser Arguments"""

    def tearDown(self):
        """Setup"""
        for handler in LOGGER.handlers:
            LOGGER.removeHandler(handler)

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_mix_in(self, yaml_mock):
        """Test mix in"""
        yaml_mock.side_effect = [
            {"include_checks": ["I", "I1111"], "regions": ["us-west-2"]},
            {},
        ]

        config = cfnlint.config.ConfigMixIn(["--regions", "us-west-1"])
        self.assertEqual(config.regions, ["us-west-1"])
        self.assertEqual(config.include_checks, ["W", "E", "I", "I1111"])

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_precedence(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [
            {
                "include_checks": ["I"],
                "ignore_checks": ["E3001"],
                "regions": ["us-west-2"],
            },
            {},
        ]
        config = cfnlint.config.ConfigMixIn(["--include-checks", "I1234", "I4321"])
        config.template_args = {
            "Metadata": {
                "cfn-lint": {
                    "config": {"include_checks": ["I9876"], "ignore_checks": ["W3001"]}
                }
            }
        }
        # config files wins
        self.assertEqual(config.regions, ["us-west-2"])
        # CLI should win
        self.assertEqual(config.include_checks, ["W", "E", "I1234", "I4321"])
        # template file wins over config file
        self.assertEqual(config.ignore_checks, ["W3001"])

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_file_output(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [{"output_file": "test_output.txt"}, {}]
        config = cfnlint.config.ConfigMixIn([])

        # Config file wins
        self.assertEqual(config.output_file, "test_output.txt")

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_file_output_mixin(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [{"output_file": "test_output.txt"}, {}]
        config = cfnlint.config.ConfigMixIn(["--output-file", "test_output_2.txt"])

        # CLI args win
        self.assertEqual(config.output_file, "test_output_2.txt")

    @patch.dict(
        os.environ,
        {
            "HOME": os.environ.get("HOME", ""),
            "USERPROFILE": os.environ.get("USERPROFILE", ""),
            "HOMEPATH": os.environ.get("HOMEPATH", ""),
            "HOMEDRIVE": os.environ.get("HOMEDRIVE", ""),
        },
        clear=True,
    )
    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_default_region(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [{}, {}]
        config = cfnlint.config.ConfigMixIn([])

        # test defaults
        self.assertEqual(config.regions, ["us-east-1"])

    @patch.dict(
        os.environ,
        {
            "AWS_REGION": "us-west-2",
            "AWS_DEFAULT_REGION": "us-west-1",
            "HOME": os.environ.get("HOME", ""),
            "USERPROFILE": os.environ.get("USERPROFILE", ""),
            "HOMEPATH": os.environ.get("HOMEPATH", ""),
            "HOMEDRIVE": os.environ.get("HOMEDRIVE", ""),
        },
        clear=True,
    )
    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_region_from_AWS_REGION_envvar(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [{}, {}]
        config = cfnlint.config.ConfigMixIn([])

        # test defaults
        self.assertEqual(config.regions, ["us-west-2"])

    @patch.dict(
        os.environ,
        {
            "AWS_DEFAULT_REGION": "us-west-1",
            "HOME": os.environ.get("HOME", ""),
            "USERPROFILE": os.environ.get("USERPROFILE", ""),
            "HOMEPATH": os.environ.get("HOMEPATH", ""),
            "HOMEDRIVE": os.environ.get("HOMEDRIVE", ""),
        },
        clear=True,
    )
    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_region_from_AWS_DEFAULT_REGION_envvar(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [{}, {}]
        config = cfnlint.config.ConfigMixIn([])

        # test defaults
        self.assertEqual(config.regions, ["us-west-1"])

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_all_regions(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [{"regions": ["ALL_REGIONS"]}, {}]
        config = cfnlint.config.ConfigMixIn([])

        # test defaults
        self.assertEqual(config.regions, REGIONS)

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_force_with_update_specs(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [{}, {}]
        config = cfnlint.config.ConfigMixIn(["--update-specs", "--force"])

        # test defaults
        self.assertEqual(config.force, True)
        self.assertEqual(config.update_specs, True)

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_expand_paths(self, yaml_mock):
        """Test precedence in"""

        yaml_mock.side_effect = [
            {"templates": ["test/fixtures/templates/public/*.yaml"]},
            {},
        ]
        config = cfnlint.config.ConfigMixIn([])

        # test defaults
        self.assertEqual(
            config.templates,
            [
                str(
                    Path(
                        "test/fixtures/templates/public"
                        + os.path.sep
                        + "lambda-poller.yaml"
                    )
                ),
                str(
                    Path(
                        "test/fixtures/templates/public"
                        + os.path.sep
                        + "rds-cluster.yaml"
                    )
                ),
            ],
        )

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_expand_paths_failure(self, yaml_mock):
        """Test precedence in"""

        filename = "test/fixtures/templates/badpath/*.yaml"
        yaml_mock.side_effect = [
            {"templates": ["test/fixtures/templates/badpath/*.yaml"]},
            {},
        ]
        config = cfnlint.config.ConfigMixIn([])

        # test defaults
        self.assertEqual(config.templates, [str(Path(filename))])

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_expand_ignore_templates(self, yaml_mock):
        """Test ignore templates"""

        yaml_mock.side_effect = [
            {
                "templates": ["test/fixtures/templates/bad/resources/iam/*.yaml"],
                "ignore_templates": [
                    "test/fixtures/templates/bad/resources/iam/resource_*.yaml"
                ],
            },
            {},
        ]
        config = cfnlint.config.ConfigMixIn([])

        # test defaults
        self.assertNotIn(
            str(Path("test/fixtures/templates/bad/resources/iam/resource_policy.yaml")),
            config.templates,
        )
        self.assertEqual(len(config.templates), 3)

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_merge(self, yaml_mock):
        """Test merging lists"""

        yaml_mock.side_effect = [
            {
                "include_checks": ["I"],
                "ignore_checks": ["E3001"],
                "regions": ["us-west-2"],
            },
            {},
        ]
        config = cfnlint.config.ConfigMixIn(
            ["--include-checks", "I1234", "I4321", "--merge-configs"]
        )
        config.template_args = {
            "Metadata": {
                "cfn-lint": {
                    "config": {"include_checks": ["I9876"], "ignore_checks": ["W3001"]}
                }
            }
        }
        # config files wins
        self.assertEqual(config.regions, ["us-west-2"])
        # CLI should win
        self.assertEqual(
            config.include_checks, ["W", "E", "I1234", "I4321", "I9876", "I"]
        )
        # template file wins over config file
        self.assertEqual(config.ignore_checks, ["W3001", "E3001"])



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.UniqueNames import UniqueNames  # pylint: disable=E0401


class TestUniqueNames(BaseRuleTestCase):
    def setUp(self):
        super(TestUniqueNames, self).setUp()
        self.collection.register(UniqueNames())

    def test_file_positive(self):
        self.helper_file_positive()

    def test_file_negative(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/uniqueNames.yaml", 1
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from io import StringIO
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

import cfnlint.decode.cfn_json  # pylint: disable=E0401
from cfnlint import ConfigMixIn
from cfnlint.config import _DEFAULT_RULESDIR
from cfnlint.rules import Rules
from cfnlint.template.template import Template  # pylint: disable=E0401


class TestCfnJson(BaseTestCase):
    """Test JSON Parsing"""

    def setUp(self):
        """SetUp template object"""
        self.rules = Rules.create_from_directory(_DEFAULT_RULESDIR)
        self.config = ConfigMixIn(
            include_experimental=True,
        )

        self.filenames = {
            "config_rule": {
                "filename": "test/fixtures/templates/quickstart/config-rules.json",
                "failures": 5,
            },
            "iam": {
                "filename": "test/fixtures/templates/quickstart/iam.json",
                "failures": 5,
            },
            "nat_instance": {
                "filename": "test/fixtures/templates/quickstart/nat-instance.json",
                "failures": 3,
            },
            "vpc_management": {
                "filename": "test/fixtures/templates/quickstart/vpc-management.json",
                "failures": 9,
            },
            "vpc": {
                "filename": "test/fixtures/templates/quickstart/vpc.json",
                "failures": 1,
            },
            "poller": {
                "filename": "test/fixtures/templates/public/lambda-poller.json",
                "failures": 1,
            },
        }

    def test_success_parse(self):
        """Test Successful JSON Parsing"""
        for _, values in self.filenames.items():
            filename = values.get("filename")
            failures = values.get("failures")

            template = cfnlint.decode.cfn_json.load(filename)
            cfn = Template(filename, template, ["us-east-1"])

            matches = list(self.rules.run(filename, cfn, self.config))
            assert (
                len(matches) == failures
            ), "Expected {} failures, got {} on {}".format(failures, matches, filename)

    def test_success_escape_character(self):
        """Test Successful JSON Parsing"""
        failures = 1
        filename = "test/fixtures/templates/good/decode/parsing.json"
        template = cfnlint.decode.cfn_json.load(filename)
        cfn = Template(filename, template, ["us-east-1"])

        matches = list(self.rules.run(filename, cfn, self.config))
        assert len(matches) == failures, "Expected {} failures, got {} on {}".format(
            failures, matches, filename
        )

    def test_success_parse_stdin(self):
        """Test Successful JSON Parsing through stdin"""
        for _, values in self.filenames.items():
            filename = None
            failures = values.get("failures")
            with open(values.get("filename"), "r") as fp:
                file_content = fp.read()
            with patch("sys.stdin", StringIO(file_content)):
                template = cfnlint.decode.cfn_json.load(filename)
                cfn = Template(filename, template, ["us-east-1"])

                matches = list(self.rules.run(filename, cfn, self.config))
                assert (
                    len(matches) == failures
                ), "Expected {} failures, got {} on {}".format(
                    failures, matches, values.get("filename")
                )

    def test_fail_run(self):
        """Test failure run"""

        filename = "test/fixtures/templates/bad/json_parse.json"

        try:
            cfnlint.decode.cfn_json.load(filename)
        except cfnlint.decode.cfn_json.JSONDecodeError:
            assert True
            return

        assert False



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.updatepolicy.Configuration import Configuration


@pytest.fixture(scope="module")
def rule():
    rule = Configuration()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        ("Valid with a non updatable type", {"Type": "AWS::Foo::Bar"}, []),
        (
            "Invalid with autoscaling group wrong type",
            {"Type": "AWS::AutoScaling::AutoScalingGroup", "UpdatePolicy": []},
            [
                ValidationError(
                    "[] is not of type 'object'",
                    path=["UpdatePolicy"],
                    rule=Configuration(),
                    instance=[],
                    validator="type",
                    validator_value="object",
                    schema_path=[
                        "allOf",
                        0,
                        "then",
                        "properties",
                        "UpdatePolicy",
                        "type",
                    ],
                )
            ],
        ),
        (
            "Invalid with autoscaling group wrong property",
            {
                "Type": "AWS::AutoScaling::AutoScalingGroup",
                "UpdatePolicy": {"Foo": "Bar"},
            },
            [
                ValidationError(
                    "Additional properties are not allowed ('Foo' was unexpected)",
                    path=["UpdatePolicy", "Foo"],
                    rule=Configuration(),
                    instance={"Foo": "Bar"},
                    validator="additionalProperties",
                    validator_value=False,
                    schema_path=[
                        "allOf",
                        0,
                        "then",
                        "properties",
                        "UpdatePolicy",
                        "additionalProperties",
                    ],
                )
            ],
        ),
        (
            "Valid with autoscaling group",
            {
                "Type": "AWS::AutoScaling::AutoScalingGroup",
                "UpdatePolicy": {
                    "AutoScalingReplacingUpdate": {
                        "WillReplace": True,
                    },
                    "AutoScalingRollingUpdate": {
                        "MaxBatchSize": 1,
                        "MinInstancesInService": "1",  # also allows strings
                        "MinSuccessfulInstancesPercent": 100,
                        "PauseTime": "PT1M",
                        "WaitOnResourceSignals": True,
                        "SuspendProcesses": ["AZRebalance"],
                    },
                    "AutoScalingScheduledAction": {
                        "IgnoreUnmodifiedGroupSizeProperties": True,
                    },
                },
            },
            [],
        ),
        (
            "Valid with lambda function",
            {
                "Type": "AWS::Lambda::Alias",
                "UpdatePolicy": {
                    "CodeDeployLambdaAliasUpdate": {
                        "AfterAllowTrafficHook": "Foo",
                        "ApplicationName": "Foo",
                        "BeforeAllowTrafficHook": "Foo",
                        "DeploymentGroupName": "Foo",
                    }
                },
            },
            [],
        ),
        (
            "Valid with search resource",
            {
                "Type": "AWS::Elasticsearch::Domain",
                "UpdatePolicy": {
                    "EnableVersionUpgrade": True,
                },
            },
            [],
        ),
        (
            "Valid with ElastiCache resource",
            {
                "Type": "AWS::ElastiCache::ReplicationGroup",
                "UpdatePolicy": {"UseOnlineResharding": True},
            },
            [],
        ),
    ],
)
def test_update_policy_configuration(name, instance, expected, rule, validator):
    errors = list(rule.validate(validator, {}, instance, {}))
    assert errors == expected, f"Test {name!r} got {errors!r}"



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any, List, Tuple

from cfnlint.match import Match

TransformResult = Tuple[List[Match], Any]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.metadata.InterfaceParameterExists import InterfaceParameterExists


@pytest.fixture(scope="module")
def rule():
    rule = InterfaceParameterExists()
    yield rule


@pytest.fixture
def template():
    return {
        "Parameters": {
            "Foo": {"Type": "String"},
        },
    }


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid Interface",
            "Foo",
            [],
        ),
        (
            "Wrong parameter",
            "Bar",
            [
                ValidationError(
                    ("'Bar' is not one of ['Foo']"),
                    validator="enum",
                    schema_path=deque(["enum"]),
                    rule=InterfaceParameterExists(),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



[
    {
        "op": "test",
        "path": "/createOnlyProperties",
        "value": [
            "/properties/Tags",
            "/properties/ExperimentOptions/properties/AccountTargeting",
        ],
    },
    {
        "op": "replace",
        "path": "/createOnlyProperties",
        "value": ["/properties/Tags", "/properties/ExperimentOptions/AccountTargeting"],
    },
]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.parameters.Default import Default  # pylint: disable=E0401


class TestDefault(BaseRuleTestCase):
    """Test template parameter configurations"""

    def setUp(self):
        """Setup"""
        super(TestDefault, self).setUp()
        self.collection.register(Default())

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_parameters_success(self):
        self.helper_file_positive_template(
            "test/fixtures/templates/good/parameters/default.yaml"
        )

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/parameters/default.yaml", 9
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.cloudformation.NestedStackParameters import (
    NestedStackParameters,  # pylint: disable=E0401
)


class TestNestedStackParameters(BaseRuleTestCase):
    """Test CloudFormation Nested stack parameters"""

    def tearDown(self) -> None:
        super().tearDown()
        logger = logging.getLogger("cfnlint.decode.decode")
        logger.disabled = False

    def setUp(self):
        """Setup"""
        super(TestNestedStackParameters, self).setUp()
        self.collection.register(NestedStackParameters())
        logger = logging.getLogger("cfnlint.decode.decode")
        logger.disabled = True
        self.success_templates = [
            "test/fixtures/templates/good/resources/cloudformation/stacks.yaml"
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        err_count = 8
        self.helper_file_negative(
            "test/fixtures/templates/bad/resources/cloudformation/stacks.yaml",
            err_count,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules.resources.iam.Policy import Policy


class IdentityPolicy(Policy):
    """Check IAM identity Policies"""

    id = "E3510"
    shortdesc = "Validate identity based IAM polices"
    description = (
        "IAM identity polices are embedded JSON in CloudFormation. "
        "This rule validates those embedded policies."
    )
    source_url = "https://docs.aws.amazon.com/IAM/latest/UserGuide/access_policies_identity-vs-resource.html"
    tags = ["resources", "iam"]

    def __init__(self):
        super().__init__(
            [
                "Resources/AWS::IAM::Group/Properties/Policies/*/PolicyDocument",
                "Resources/AWS::IAM::ManagedPolicy/Properties/PolicyDocument",
                "Resources/AWS::IAM::Policy/Properties/PolicyDocument",
                "Resources/AWS::IAM::Role/Properties/Policies/*/PolicyDocument",
                "Resources/AWS::IAM::User/Properties/Policies/*/PolicyDocument",
                "Resources/AWS::SSO::PermissionSet/Properties/InlinePolicy",
            ],
            "identity",
            "policy_identity.json",
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import subprocess
import unittest
from pathlib import Path
from typing import Any, Dict, List

from cfnlint.config import configure_logging
from cfnlint.decode import cfn_yaml
from cfnlint.runner import Runner


class BaseCliTestCase(unittest.TestCase):
    """Used for Testing CLI"""

    scenarios: List[Dict[str, Any]] = []

    def setUp(self):
        """Common Settings"""
        self.maxDiff = None

    def run_scenarios(self, extra_params=None):
        """Success test"""
        extra_params = extra_params or []
        for scenario in self.scenarios:
            filename = scenario.get("filename")
            results_filename = scenario.get("results_filename")
            expected_results = scenario.get("results", [])
            exit_code = scenario.get("exit_code")

            if results_filename and not expected_results:
                with open(results_filename, encoding="utf-8") as json_data:
                    expected_results = json.load(json_data)

            for result in expected_results:
                result["Filename"] = str(Path(result.get("Filename")))

            try:
                result = subprocess.check_output(
                    ["cfn-lint"] + extra_params + ["--format", "json", "--", filename]
                )

                self.assertEqual(
                    0,
                    exit_code,
                    "Expected {} exit code, got {} on {}".format(
                        exit_code, 0, filename
                    ),
                )

                if isinstance(result, bytes):
                    result = result.decode("utf8")
                matches = json.loads(result)

                self.assertCountEqual(
                    expected_results,
                    matches,
                    "Expected {} failures, got {} on {}".format(
                        len(expected_results), len(matches), filename
                    ),
                )

            except subprocess.CalledProcessError as error:
                self.assertEqual(
                    error.returncode,
                    exit_code,
                    "Expected {} exit code, got {} on {}".format(
                        exit_code, error.returncode, filename
                    ),
                )

                if isinstance(error.output, bytes):
                    error.output = error.output.decode("utf8")
                matches = json.loads(error.output)

                self.assertCountEqual(
                    expected_results,
                    matches,
                    "Expected {} failures, got {} on {}".format(
                        len(expected_results), len(matches), filename
                    ),
                )

    def run_module_integration_scenarios(self, config):
        """Test using cfnlint as a module integrated into another package"""

        configure_logging(None, False)

        for scenario in self.scenarios:
            filename = scenario.get("filename")
            results_filename = scenario.get("results_filename")
            expected_results = scenario.get("results", [])

            if results_filename and not expected_results:
                with open(results_filename, encoding="utf-8") as json_data:
                    expected_results = json.load(json_data)

            for result in expected_results:
                result["Filename"] = str(Path(result.get("Filename")))

            template = cfn_yaml.load(filename)

            runner = Runner(config)
            matches = list(runner.validate_template(filename, template))

            # Only check that the error count matches as the formats are different
            self.assertEqual(
                len(expected_results),
                len(matches),
                "Expected {} failures, got {} on {}".format(
                    len(expected_results), matches, filename
                ),
            )
